"""String serialization."""
from __future__ import annotations
import re
import math
from .. import util
from .. import algebra as alg
from .color_names import to_name
from ..channels import FLG_PERCENT, FLG_OPT_PERCENT, FLG_ANGLE
from ..types import Vector
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color

RE_COMPRESS = re.compile(r'(?i)^#([a-f0-9])\1([a-f0-9])\2([a-f0-9])\3(?:([a-f0-9])\4)?$')

COMMA = ', '
SLASH = ' / '
SPACE = ' '
EMPTY = ''


def named_color(
    obj: 'Color',
    alpha: bool | None,
    fit: str | bool
) -> str | None:
    """Get the CSS color name."""

    a = get_alpha(obj, alpha, False, False)
    if a is None:
        a = 1
    return to_name(get_coords(obj, fit, False, False) + [a])


def named_color_function(
    obj: 'Color',
    func: str,
    alpha: bool | None,
    precision: int,
    fit: str | bool,
    none: bool,
    percent: bool,
    legacy: bool,
    scale: float
) -> str:
    """Translate to CSS function form `name(...)`."""

    # Create the function `name` or `namea` if old legacy form.
    a = get_alpha(obj, alpha, none, legacy)
    string = ['{}{}('.format(func, 'a' if legacy and a is not None else EMPTY)]

    # Iterate the coordinates formatting them for percent, not percent, and even scaling them (sRGB).
    coords = get_coords(obj, fit, none, legacy)
    channels = obj._space.CHANNELS
    for idx, value in enumerate(coords):
        channel = channels[idx]
        use_percent = channel.flags & FLG_PERCENT or (percent and channel.flags & FLG_OPT_PERCENT)
        is_angle = channel.flags & FLG_ANGLE
        if not use_percent and not is_angle:
            value *= scale
        if idx != 0:
            string.append(COMMA if legacy else SPACE)
        string.append(
            util.fmt_float(
                value,
                precision,
                channel.span if use_percent else 0.0,
                channel.offset if use_percent else 0.0
            )
        )

    # Add alpha if needed
    if a is not None:
        string.append('{}{})'.format(COMMA if legacy else SLASH, util.fmt_float(a, max(precision, util.DEF_PREC))))
    else:
        string.append(')')
    return EMPTY.join(string)


def color_function(
    obj: 'Color',
    alpha: bool | None,
    precision: int,
    fit: str | bool,
    none: bool
) -> str:
    """Color format."""

    # Export in the `color(space ...)` format
    coords = get_coords(obj, fit, none, False)
    a = get_alpha(obj, alpha, none, False)
    return (
        'color({} {}{})'.format(
            obj._space._serialize()[0],
            SPACE.join([util.fmt_float(coord, precision) for coord in coords]),
            SLASH + util.fmt_float(a, max(precision, util.DEF_PREC)) if a is not None else EMPTY
        )
    )


def get_coords(
    obj: 'Color',
    fit: str | bool,
    none: bool,
    legacy: bool
) -> Vector:
    """Get the coordinates."""

    color = (obj.fit(method=None if not isinstance(fit, str) else fit) if fit else obj)
    return color.coords(nans=False if legacy or not none else True)


def get_alpha(
    obj: 'Color',
    alpha: bool | None,
    none: bool,
    legacy: bool
) -> float | None:
    """Get the alpha if required."""

    a = obj.alpha(nans=False if not none or legacy else True)
    alpha = alpha is not False and (alpha is True or a < 1.0 or math.isnan(a))
    return None if not alpha else a


def hexadecimal(
    obj: 'Color',
    alpha: bool | None = None,
    fit: str | bool = True,
    upper: bool = False,
    compress: bool = False
) -> str:
    """Get the hex `RGB` value."""

    coords = get_coords(obj, fit, False, False)
    a = get_alpha(obj, alpha, False, False)

    if a is not None:
        value = "#{:02x}{:02x}{:02x}{:02x}".format(
            int(alg.round_half_up(coords[0] * 255.0)),
            int(alg.round_half_up(coords[1] * 255.0)),
            int(alg.round_half_up(coords[2] * 255.0)),
            int(alg.round_half_up(a * 255.0))
        )
    else:
        value = "#{:02x}{:02x}{:02x}".format(
            int(alg.round_half_up(coords[0] * 255.0)),
            int(alg.round_half_up(coords[1] * 255.0)),
            int(alg.round_half_up(coords[2] * 255.0))
        )

    if upper:
        value = value.upper()

    if compress:
        m = RE_COMPRESS.match(value)
        return (m.expand(r"#\1\2\3\4") if len(value) == 9 else m.expand(r"#\1\2\3")) if m is not None else value
    else:
        return value


def serialize_css(
    obj: 'Color',
    func: str = '',
    color: bool = False,
    alpha: bool | None = None,
    precision: int | None = None,
    fit: str | bool = True,
    none: bool = False,
    percent: bool = False,
    hexa: bool = False,
    upper: bool = False,
    compress: bool = False,
    name: bool = False,
    legacy: bool = False,
    scale: float = 1.0
) -> str:
    """Convert color to CSS."""

    if precision is None:
        precision = obj.PRECISION

    # Color format
    if color:
        return color_function(obj, alpha, precision, fit, none)

    # CSS color names
    if name:
        n = named_color(obj, alpha, fit)
        if n is not None:
            return n

    # Hex RGB
    if hexa:
        return hexadecimal(obj, alpha, fit, upper, compress)

    # Normal CSS named function format
    if func:
        return named_color_function(obj, func, alpha, precision, fit, none, percent, legacy, scale)

    raise RuntimeError('Could not identify a CSS format to serialize to')  # pragma: no cover



"""
Delta E CAM16.

https://www.ncbi.nlm.nih.gov/pmc/articles/PMC9698626/pdf/sensors-22-08869.pdf
"""
from __future__ import annotations
import math
from ..distance import DeltaE
from typing import Any, TYPE_CHECKING
from ..spaces.cam16_ucs import COEFFICENTS

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class DECAM16(DeltaE):
    """Delta E CAM16 class."""

    NAME = "cam16"

    def distance(self, color: Color, sample: Color, model: str = 'ucs', **kwargs: Any) -> float:
        """Delta E z color distance formula."""

        space = 'cam16-{}'.format(model)
        kl = COEFFICENTS[model][0]
        j1, a1, b1 = color.convert(space).coords(nans=False)
        j2, a2, b2 = sample.convert(space).coords(nans=False)

        dj = j1 - j2
        da = a1 - a2
        db = b1 - b2

        return math.sqrt((dj / kl) ** 2 + da ** 2 + db ** 2)



"""Test OkLCh library."""
import unittest
from . import util
from coloraide import NaN
from coloraide import Color
import pytest


class TestsOkLCh(util.ColorAssertsPyTest):
    """Test OkLCh."""

    COLORS = [
        # Test general conversion
        ('red', 'color(--oklch 0.62796 0.25768 29.234)'),
        ('orange', 'color(--oklch 0.79269 0.17103 70.67)'),
        ('yellow', 'color(--oklch 0.96798 0.21101 109.77)'),
        ('green', 'color(--oklch 0.51975 0.17686 142.5)'),
        ('blue', 'color(--oklch 0.45201 0.31321 264.05)'),
        ('indigo', 'color(--oklch 0.33898 0.17927 301.68)'),
        ('violet', 'color(--oklch 0.7619 0.18612 327.21)'),
        ('white', 'color(--oklch 1 0 none)'),
        ('gray', 'color(--oklch 0.59987 0 none)'),
        ('black', 'color(--oklch 0 0 none)'),
        # Test CSS
        ('oklch(1 0.3 270)', 'color(--oklch 1 0.3 270 / 1)'),
        ('oklch(1 0.3 270 / 0.5)', 'color(--oklch 1 0.3 270 / 0.5)'),
        ('oklch(50% 0.3 270)', 'color(--oklch 0.5 0.3 270 / 1)'),
        ('oklch(50% 50% 270 / 50%)', 'color(--oklch 0.5 0.2 270 / 0.5)'),
        ('oklch(none none none / none)', 'color(--oklch none none none / none)'),
        ('oklch(1 0.3 50%)', None),
        ('oklch(1, 0.3, 50)', None),
        # Test degrees
        ('oklch(0.75 0.2 180deg)', 'color(--oklch 0.75 0.2 180 / 1)'),
        ('oklch(0.75 0.2 0.5turn)', 'color(--oklch 0.75 0.2 180 / 1)'),
        ('oklch(0.75 0.2 3.14159rad)', 'color(--oklch 0.75 0.2 180 / 1)'),
        ('oklch(0.75 0.2 200grad)', 'color(--oklch 0.75 0.2 180 / 1)'),
        # Test color
        ('color(--oklch 1 0.3 270)', 'color(--oklch 1 0.3 270)'),
        ('color(--oklch 1 0.3 270 / 0.5)', 'color(--oklch 1 0.3 270 / 0.5)'),
        ('color(--oklch 50% 50% 50% / 50%)', 'color(--oklch 0.5 0.2 180 / 0.5)'),
        ('color(--oklch none none none / none)', 'color(--oklch none none none / none)'),
        # Test percent ranges
        ('color(--oklch 0% 0% 0%)', 'color(--oklch 0 0 0)'),
        ('color(--oklch 100% 100% 100%)', 'color(--oklch 1 0.4 360 / 1)'),
        ('color(--oklch -100% -100% -100%)', 'color(--oklch -1 0 -360 / 1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        if color2 is None:
            with pytest.raises(ValueError):
                Color(color1)
        else:
            self.assertColorEqual(Color(color1).convert('oklch'), Color(color2), color=True)


class TestOkLChSerialize(util.ColorAssertsPyTest):
    """Test OkLCh serialization."""

    COLORS = [
        # Test hex no options
        ('oklch(0.75 0.3 50)', {}, 'oklch(0.75 0.3 50)'),
        # Test alpha
        ('oklch(0.75 0.3 50 / 0.5)', {}, 'oklch(0.75 0.3 50 / 0.5)'),
        ('oklch(0.75 0.3 50)', {'alpha': True}, 'oklch(0.75 0.3 50 / 1)'),
        ('oklch(0.75 0.3 50 / 0.5)', {'alpha': False}, 'oklch(0.75 0.3 50)'),
        # Test percent
        ('oklch(0.5 0.2 180)', {'percent': True}, 'oklch(50% 50% 180)'),
        ('oklch(0.5 0.2 180 / 0.5)', {'percent': True, 'alpha': True}, 'oklch(50% 50% 180 / 0.5)'),
        # Test None
        ('oklch(none 0.3 50)', {}, 'oklch(0 0.3 50)'),
        ('oklch(none 0.3 50)', {'none': True}, 'oklch(none 0.3 50)'),
        # Test fit (not bound)
        ('oklch(0.2 0.5 0)', {}, 'oklch(0.2 0.5 0)'),
        ('oklch(0.2 0.5 0)', {'fit': False}, 'oklch(0.2 0.5 0)'),
        # Test color
        ('oklch(none 0.3 50 / 0.5)', {'color': True}, 'color(--oklch 0 0.3 50 / 0.5)'),
        ('oklch(none 0.3 50)', {'color': True, 'none': True}, 'color(--oklch none 0.3 50)'),
        ('oklch(0 0.3 50)', {'color': True, 'alpha': True}, 'color(--oklch 0 0.3 50 / 1)'),
        ('oklch(0 0.3 50 / 0.5)', {'color': True, 'alpha': False}, 'color(--oklch 0 0.3 50)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestOkLChProperties(util.ColorAsserts, unittest.TestCase):
    """Test OkLCh."""

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--oklch 0.9 0.5 270 / 1)')
        self.assertEqual(c['lightness'], 0.9)
        c['lightness'] = 0.2
        self.assertEqual(c['lightness'], 0.2)

    def test_chroma(self):
        """Test `chroma`."""

        c = Color('color(--oklch 0.9 0.5 270 / 1)')
        self.assertEqual(c['chroma'], 0.5)
        c['chroma'] = 0.1
        self.assertEqual(c['chroma'], 0.1)

    def test_hue(self):
        """Test `hue`."""

        c = Color('color(--oklch 0.9 0.5 270 / 1)')
        self.assertEqual(c['hue'], 270)
        c['hue'] = 0.1
        self.assertEqual(c['hue'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--oklch 0.9 0.5 270 / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('oklch', [0.9, 50, NaN], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_none_input(self):
        """Test `none` null."""

        c = Color('color(--oklch 90% 0 none / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_near_zero_null(self):
        """
        Test very near zero null.

        This is a fix up to help give more sane hues
        when chroma is very close to zero.
        """

        c = Color('color(--oklch 90% 0.00000009 120 / 1)').convert('srgb').convert('oklch')
        self.assertTrue(c.is_nan('hue'))

    def test_null_normalization_min_chroma(self):
        """Test minimum saturation."""

        c = Color('oklch(90% 0 120 / 1)').normalize()
        self.assertTrue(c.is_nan('hue'))

    def test_from_oklab(self):
        """Test null from Lab conversion."""

        c1 = Color('color(--oklab 90% 0 0)')
        c2 = c1.convert('oklch')
        self.assertColorEqual(c2, Color('color(--oklch 90% 0 none)'))
        self.assertTrue(c2.is_nan('hue'))

    def test_achromatic_hue(self):
        """Test that all RGB-ish colors convert to OkLCh with a null hue."""

        for space in ('srgb', 'display-p3', 'rec2020', 'a98-rgb', 'prophoto-rgb'):
            for x in range(0, 256):
                color = Color('color({space} {num:f} {num:f} {num:f})'.format(space=space, num=x / 255))
                color2 = color.convert('oklch')
                self.assertTrue(color2.is_nan('hue'))


class TestQuirks(util.ColorAsserts, unittest.TestCase):
    """Test quirks."""

    def test_chorma_less_than_zero_convert(self):
        """Test handling of negative chroma when converting to Oklab."""

        self.assertColorEqual(Color('color(--oklch 90% -10 120 / 1)').convert('oklab'), Color('color(--oklab 0.9 0 0)'))


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('oklch', [0.3, 0, 270]).is_achromatic(), True)
        self.assertEqual(Color('oklch', [0.3, 0.000001, 270]).is_achromatic(), True)
        self.assertEqual(Color('oklch', [NaN, 0.00001, 270]).is_achromatic(), True)
        self.assertEqual(Color('oklch', [0, NaN, 270]).is_achromatic(), True)
        self.assertEqual(Color('oklch', [0, 1, 270]).is_achromatic(), False)
        self.assertEqual(Color('oklch', [NaN, 0.2, 270]).is_achromatic(), False)
        self.assertEqual(Color('oklch', [0.3, NaN, 270]).is_achromatic(), True)
        self.assertEqual(Color('oklch', [NaN, NaN, 270]).is_achromatic(), True)



"""Apply compositing on images of the same size."""
from functools import lru_cache
from PIL import Image
import time
import argparse
import sys
import os

# We want to load ColorAide from the working directory to pick up the version under development
sys.path.insert(0, os.getcwd())

try:
    from coloraide_extras.everything import ColorAll as Color
except ImportError:
    from coloraide.everything import ColorAll as Color


def printt(t):
    """Print time."""

    print('Completed in: ', end='')
    s = t // 1e+9
    m = t // 1e+6
    u = t // 1000
    if s:
        s = t / 1e+9
        h = m = 0
        m = s // 60
        if m:
            s -= m * 60
            h = m // 60
            if h:
                m -= h * 60
        if h:
            print('{} hours '.format(h), end='')
        if m:
            print('{} minutes '.format(m), end='')
        print('{} sec'.format(s))
    elif m:
        print('{} msec'.format(t / 1e+6))
    elif u:
        print('{} usec'.format(t / 1000))
    else:
        print('{} nsec'.format(t))


@lru_cache(maxsize=1024 * 1024)
def apply_compositing(background, blend, operator, pixels, fit, space):
    """Apply compositing."""

    color = Color('srgb', [x / 255 for x in pixels[0][:-1]], pixels[0][3] / 255)
    backdrops = [Color('srgb', [x / 255 for x in p[:3]], p[3] / 255) for p in pixels[1:]]
    color.compose(backdrops, blend=blend, operator=operator, in_place=True, space=space, out_space='srgb').clip()
    # Overlay first layer on background color in isolation
    if background != 'transparent':
        color.compose(background, in_place=True).clip()
    return tuple(int(x * 255) for x in color.fit(method=fit)[:4])


def process_image(imgs, bg, output, blend, porter_duff, fit, space):
    """Process the image applying the requested blend mode and compositing operator."""

    images = []
    pixels = []
    size = None
    for i in imgs:
        img = Image.open(i)

        # Ensure we are working in `RGBA`
        if img.mode not in ('RGBA',):
            img = img.convert('RGBA')

        if size is None:
            size = img.size
        elif size != img.size:
            raise RuntimeError('Images must be of the same size')

        images.append(img)
        pixels.append(img.load())

    start = time.perf_counter_ns()
    x, y = images[0].size
    total = images[0].size[0] * images[0].size[1]
    factor = 100 / total

    # Compose the various layers
    i = j = 0
    print('Pixels: {}'.format(total))
    print('> 0%', end='\r')
    for i in range(x):
        for j in range(y):
            pixels[0][i, j] = apply_compositing(bg, blend, porter_duff, tuple([p[i, j] for p in pixels]), fit, space)
        print('> {}%'.format(int((i * j) * factor)), end="\r")
    print('> 100%')
    t = time.perf_counter_ns() - start
    printt(t)
    images[0].save(output)

    # Close images
    [img.close() for img in images]


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='compose_images.py', description='Apply compositing to a series of images.')
    parser.add_argument('--image', '-i', action='append', help='Image.')
    parser.add_argument('--output', '-o', help='Output file.')
    parser.add_argument('--porter-duff', '-p', default='source-over', help='Porter Duff compositing operator.')
    parser.add_argument('--blend', '-b', default='normal', help="The blend mode.")
    parser.add_argument('--space', '-s', default='srgb-linear', help="Space to compose in.")
    parser.add_argument(
        '--background-color', '-c',
        default='transparent',
        help="Background color. Blended in source-over in isolation."
    )
    parser.add_argument('--gamut-map', '-g', default="clip", help="Specify GMA method to use (default simple clipping)")
    args = parser.parse_args()

    process_image(
        args.image,
        args.background_color,
        args.output,
        args.blend,
        args.porter_duff,
        args.gamut_map,
        args.space
    )

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Test plugins and general configuration."""
import unittest
from coloraide import Color
from coloraide import util as cutil
from . import util


class TestCustom(util.ColorAsserts, unittest.TestCase):
    """Test custom overrides and plugins."""

    def test_override_precision(self):
        """Test precision override."""

        class Color2(Color):
            """Color."""

            PRECISION = 3

        self.assertEqual(
            Color('color(srgb 0.1234567 0.1234567 0.1234567)').to_string(color=True),
            'color(srgb 0.12346 0.12346 0.12346)'
        )

        self.assertEqual(
            Color2('color(srgb 0.1234567 0.1234567 0.1234567)').to_string(color=True),
            'color(srgb 0.123 0.123 0.123)'
        )

    def test_override_fit(self):
        """Test fit override."""

        class Color2(Color):
            """Color."""

            FIT = "clip"

        self.assertEqual(
            Color('color(srgb 2 -1 0)').fit().to_string(),
            'rgb(255 101.26 98.291)'
        )

        self.assertEqual(
            Color2('color(srgb 2 -1 0)').fit().to_string(),
            'rgb(255 0 0)'
        )

    def test_override_interpolate(self):
        """Test interpolation override."""

        class Color2(Color):
            """Color."""

            INTERPOLATE = "lab"

        self.assertEqual(
            Color('red').mix('green').to_string(),
            Color('red').mix('green', space=cutil.DEF_INTERPOLATE).to_string()
        )
        self.assertEqual(
            Color2('red').mix('green').to_string(),
            Color2('red').mix('green', space='lab').to_string()
        )

    def test_override_harmony(self):
        """Test harmony override."""

        class Color2(Color):
            """Color."""

            HARMONY = "hsl"

        self.assertEqual(
            Color('red').harmony('complement'),
            Color('red').harmony('complement', space=Color.HARMONY)
        )
        self.assertEqual(
            Color2('red').harmony('complement'),
            Color2('red').harmony('complement', space='hsl')
        )

    def test_override_delta_e(self):
        """Test delta e override."""

        class Color2(Color):
            """Color."""

            DELTA_E = "2000"

        self.assertCompare(
            Color('red').delta_e("blue"),
            176.3085
        )

        self.assertCompare(
            Color2('red').delta_e("blue"),
            52.8782
        )

    def test_plugin_registration_space(self):
        """Test plugin registration of `Space`."""

        from coloraide.spaces import lab_d65

        expected = Color('red').convert('lab-d65').to_string()

        class Custom(Color):
            pass

        Custom.deregister('space:lab-d65')
        # Deregistration should have taken place
        with self.assertRaises(ValueError):
            Custom('red').convert('lab-d65')

        # But it should not affect the base class
        self.assertEqual(Color('red').convert('lab-d65').to_string(), expected)

        # Now it is registered again
        Custom.register(lab_d65.LabD65())
        self.assertEqual(Custom('red').convert('lab-d65').to_string(), expected)

    def test_plugin_registration_delta_e(self):
        """Test plugin registration of `DeltaE`."""

        from coloraide.distance import delta_e_2000

        expected = Color('red').delta_e('green', method='2000')

        class Custom(Color):
            pass

        Custom.deregister('delta-e:2000')
        # Deregistration should have taken place
        with self.assertRaises(ValueError):
            Custom('red').delta_e('green', method='2000')

        # But it should not affect the base class
        self.assertEqual(Color('red').delta_e('green', method='2000'), expected)

        # Now it is registered again
        Custom.register(delta_e_2000.DE2000())
        self.assertEqual(Custom('red').delta_e('green', method='2000'), expected)

    def test_plugin_registration_fit(self):
        """Test plugin registration of `Fit`."""

        from coloraide.gamut import fit_lch_chroma

        expected = Color('color(srgb 110% 140% 20%)').fit(method='lch-chroma').to_string()

        class Custom(Color):
            pass

        Custom.deregister('fit:lch-chroma')
        # Deregistration should have taken place
        with self.assertRaises(ValueError):
            Custom('color(srgb 110% 140% 20%)').fit(method='lch-chroma')

        # But it should not affect the base class
        self.assertEqual(Color('color(srgb 110% 140% 20%)').fit(method='lch-chroma').to_string(), expected)

        # Now it is registered again
        Custom.register(fit_lch_chroma.LChChroma())
        self.assertEqual(Custom('color(srgb 110% 140% 20%)').fit(method='lch-chroma').to_string(), expected)

    def test_plugin_registration_contrast(self):
        """Test plugin registration of `ColorContrast`."""

        from coloraide.contrast import wcag21

        expected = Color('red').contrast('blue', method='wcag21')

        class Custom(Color):
            pass

        Custom.deregister('contrast:wcag21')
        # Deregistration should have taken place
        with self.assertRaises(ValueError):
            Custom('red').contrast('blue', method='wcag21')

        # But it should not affect the base class
        self.assertEqual(Color('red').contrast('blue', method='wcag21'), expected)

        # Now it is registered again
        Custom.register(wcag21.WCAG21Contrast())
        self.assertEqual(Custom('red').contrast('blue', method='wcag21'), expected)

    def test_plugin_registration_cat(self):
        """Test plugin registration of `cat`."""

        from coloraide.cat import VonKries, WHITES

        expected = [0.42980066937825656, 0.21417803538899305, 0.014645639757305108]

        # Deregistration should have taken place
        class Custom(Color):
            pass

        c1 = Custom('red').convert('xyz-d65')

        # Register the plugin
        Custom.register(VonKries())
        self.assertEqual(
            Custom.chromatic_adaptation(
                WHITES['2deg']['D65'],
                WHITES['2deg']['D50'],
                c1[:-1],
                method='von-kries'
            ),
            expected
        )

        # But it should not affect the base class
        with self.assertRaises(ValueError):
            Color.chromatic_adaptation(
                WHITES['2deg']['D65'],
                WHITES['2deg']['D50'],
                c1[:-1],
                method='von-kries'
            )

        Custom.deregister('cat:von-kries')
        with self.assertRaises(ValueError):
            Custom.chromatic_adaptation(
                WHITES['2deg']['D65'],
                WHITES['2deg']['D50'],
                c1[:-1],
                method='von-kries'
            )

    def test_plugin_registration_filter(self):
        """Test plugin registration of `filter`."""

        from coloraide.filters.w3c_filter_effects import Sepia

        # Deregistration should have taken place
        class Custom(Color):
            pass

        Custom.deregister('filter:sepia')
        color = Color('red').filter('sepia')

        with self.assertRaises(ValueError):
            Custom('red').filter('sepia')

        # Now it is registered again
        Custom.register(Sepia())
        self.assertColorEqual(
            Custom('red').filter('sepia'),
            color
        )

    def test_plugin_registration_interpolate(self):
        """Test plugin registration of `interpolate`."""

        from coloraide.interpolate.bspline import BSpline

        # Deregistration should have taken place
        class Custom(Color):
            pass

        Custom.deregister('interpolate:bspline')
        color = Color('red').mix('blue', method='bspline')

        with self.assertRaises(ValueError):
            Custom('red').mix('blue', method='bspline')

        # Now it is registered again
        Custom.register(BSpline())
        self.assertColorEqual(
            Custom('red').mix('blue', method='bspline'),
            color
        )

    def test_plugin_registration_cct(self):
        """Test plugin registration of `cct`."""

        from coloraide.temperature.robertson_1968 import Robertson1968

        # Deregistration should have taken place
        class Custom(Color):
            pass

        Custom.deregister('cct:robertson-1968')
        color = Color.blackbody('srgb-linear', 5000, method='robertson-1968')

        with self.assertRaises(ValueError):
            Custom.blackbody('srgb-linear', 5000, method='robertson-1968')

        # Now it is registered again
        Custom.register(Robertson1968())
        self.assertColorEqual(
            Custom.blackbody('srgb-linear', 5000, method='robertson-1968'),
            color
        )

    def test_deregister_all_category(self):
        """Test deregistration of all plugins in a category."""

        class Custom(Color):
            pass

        Custom.deregister('fit:*')
        self.assertEqual(Custom.FIT_MAP, {})
        self.assertNotEqual(Custom.CS_MAP, {})

    def test_deregister_all(self):
        """Test deregistration of all plugins."""

        class Custom(Color):
            pass

        Custom.deregister('*')
        self.assertEqual(Custom.FIT_MAP, {})
        self.assertEqual(Custom.CS_MAP, {})
        self.assertEqual(Custom.DE_MAP, {})
        self.assertEqual(Custom.INTERPOLATE_MAP, {})
        self.assertEqual(Custom.FILTER_MAP, {})
        self.assertEqual(Custom.CAT_MAP, {})
        self.assertEqual(Custom.CONTRAST_MAP, {})
        self.assertEqual(Custom.CCT_MAP, {})

    def test_reserved_registration_fit(self):
        """Test override registration of reserved fit method."""

        from coloraide.gamut import Fit

        class Custom(Color):
            pass

        class CustomFit(Fit):
            NAME = 'clip'

            @staticmethod
            def fit(color):
                return [0, 0, 0]

        with self.assertRaises(ValueError):
            Custom.register(CustomFit(), overwrite=True)

    def test_bad_registration_type(self):
        """Test bad registration type."""

        class Custom(Color):
            pass

        class BadClass:
            pass

        with self.assertRaises(TypeError):
            Custom.register(BadClass())

    def test_bad_registration_star(self):
        """Test bad registration type."""

        from coloraide.distance import DeltaE

        class Custom(Color):
            pass

        class CustomDE(DeltaE):
            NAME = '*'

            @staticmethod
            def distance(color, sample, **kwargs):
                return 0

        with self.assertRaises(ValueError):
            Custom.register(CustomDE())

    def test_bad_registration_exists(self):
        """Test bad registration of plugin that exists."""

        from coloraide.spaces.lab_d65 import LabD65

        class Custom(Color):
            pass

        with self.assertRaises(ValueError):
            Custom.register(LabD65())

        Custom.register(LabD65(), overwrite=True)

    def test_silent_registration_exists(self):
        """Test silent handling of already registered plugin."""

        from coloraide.spaces.jzazbz import Jzazbz

        class Custom(Color):
            pass

        Custom.register(Jzazbz(), silent=True)

    def test_bad_deregister_category(self):
        """Test bad deregistration category."""

        class Custom(Color):
            pass

        with self.assertRaises(ValueError):
            Custom.deregister('bad:srgb')

    def test_bad_deregister_plugin(self):
        """Test bad deregistration category."""

        class Custom(Color):
            pass

        with self.assertRaises(ValueError):
            Custom.deregister('space:bad')

    def test_reserved_deregistration_fit(self):
        """Test deregistration of reserved fit method."""

        class Custom(Color):
            pass

        with self.assertRaises(ValueError):
            Custom.deregister('fit:clip')

    def test_silent_deregister_not_exists(self):
        """Test if a non-existing plugin is removed from a proper category, that silent will cause no error."""

        class Custom(Color):
            pass

        Custom.deregister('space:bad', silent=True)



"""Sanity check that ensures all colors round trip back."""
from coloraide.everything import ColorAll as Base
from coloraide.spaces import Cylindrical, HSLish, HSVish
from coloraide import algebra as alg
import pytest


class TestRoundTrip:
    """
    Test round trip conversions for non-achromatic colors.

    A color that does not round trip back likely has a broken conversion.
    It may be possible that some color spaces degrade through conversion.
    Only lower precision for spaces that do not have a high degree of accuracy,
    but they should round trip for their lower precision.
    """

    class Color(Base):
        """Local color object."""

    Color.deregister('space:hpluv')

    SPACES = {k: 5 for k in Color.CS_MAP.keys()}
    # Not as accurate due to bisect back to CAM16
    SPACES['hct'] = 3

    COLORS = [
        Color('red'),
        Color('orange'),
        Color('yellow'),
        Color('green'),
        Color('blue'),
        Color('indigo'),
        Color('violet')
    ]

    def assert_round_trip(self, color, space):
        """Cycle through all the other colors and convert to them and back and check the results."""

        c1 = self.Color(color)
        name = c1.to_string(names=True)
        c1.convert(space, in_place=True)
        p1 = self.SPACES[space]
        for space, p2 in self.SPACES.items():
            p = min(p1, p2)
            # Print the color space to easily identify which color space broke.
            c2 = c1.convert(space)
            c2.convert(c1.space(), in_place=True)
            # Catch cases where we are really close to 360 which should wrap to 0
            if isinstance(c2._space, Cylindrical):
                if alg.round_half_up(alg.no_nan(c2['hue']), p) == 360:
                    c2.set('hue', 0)
            # In HSL and HSV spaces particularly, we can get nonsense saturation if lightness
            # is not exactly within 0 - 1 range. Ignore saturation in achromatic cases.
            if isinstance(c2._space, (HSLish, HSVish)) and alg.is_nan(c2['hue']):
                c2[c2._space.indexes()[1]] = 0.0
            if isinstance(c1._space, (HSLish, HSVish)) and alg.is_nan(c1['hue']):
                c1[c1._space.indexes()[1]] = 0.0
            # Run rounded string back through parsing in case we hit something like a hue that needs normalization.
            str1 = self.Color(c1.to_string(color=True, fit=False)).to_string(color=True, fit=False, precision=p)
            str2 = self.Color(c2.to_string(color=True, fit=False)).to_string(color=True, fit=False, precision=p)
            # Print failing results for debug purposes
            if str1 != str2:
                print('----- Convert: {} <=> {} -----'.format(c1.space(), space))
                print('Name: ', name)
                print('Original: ', color.to_string(color=True, fit=False, precision=p))
                print(c1.space() + ': ', str1, c1[:])
                print(space + ': ', str2, c2[:])
                assert str1 == str2

    @pytest.mark.parametrize('space', SPACES)
    def test_round_trip(self, space):
        """Test round trip."""

        for c in self.COLORS:
            self.assert_round_trip(c, space)


class TestAchromaticRoundTrip(TestRoundTrip):
    """
    Test achromatic round trip.

    For convenience, we determine achromatic colors and set hues to `NaN`.
    This can cause a slight drop in precision.
    """

    class Color(Base):
        """Local color object."""

    Color.deregister('space:hpluv')

    SPACES = {k: 5 for k in Color.CS_MAP.keys()}
    # Precision just isn't as high for these in achromatic region
    # but it is good enough for practical purposes.

    COLORS = [
        Color('darkgrey'),
        Color('lightgrey'),
        Color('gray'),
        Color('black'),
        Color('white')
    ]

    @pytest.mark.parametrize('space', SPACES)
    def test_round_trip(self, space):
        """Test round trip."""

        for c in self.COLORS:
            self.assert_round_trip(c, space)



"""Test DIN99o library."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color
import pytest


class TestDIN99o(util.ColorAssertsPyTest):
    """Test DIN99o."""

    COLORS = [
        ('red', 'color(--din99o 57.289 39.498 30.518)'),
        ('orange', 'color(--din99o 77.855 16.444 40.318)'),
        ('yellow', 'color(--din99o 97.552 -8.046 44.345)'),
        ('green', 'color(--din99o 50.336 -30.169 25.591)'),
        ('blue', 'color(--din99o 36.03 31.935 -40.383)'),
        ('indigo', 'color(--din99o 23.324 29.57 -27.086)'),
        ('violet', 'color(--din99o 73.015 34.267 -18.421)'),
        ('white', 'color(--din99o 100 0 0)'),
        ('gray', 'color(--din99o 57.63 0 0)'),
        ('black', 'color(--din99o 0 0 0)'),
        # Test color
        ('color(--din99o 50 10 -10)', 'color(--din99o 50 10 -10)'),
        ('color(--din99o 50 10 -10 / 0.5)', 'color(--din99o 50 10 -10 / 0.5)'),
        ('color(--din99o 50% 50% -50% / 50%)', 'color(--din99o 50 27.5 -27.5 / 0.5)'),
        ('color(--din99o none none none / none)', 'color(--din99o none none none / none)'),
        # Test percent ranges
        ('color(--din99o 0% 0% 0%)', 'color(--din99o 0 0 0)'),
        ('color(--din99o 100% 100% 100%)', 'color(--din99o 100 55 55)'),
        ('color(--din99o -100% -100% -100%)', 'color(--din99o -100 -55 -55)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('din99o'), Color(color2))


class TestDIN99oSerialize(util.ColorAssertsPyTest):
    """Test DIN99o serialization."""

    COLORS = [
        # Test color
        ('color(--din99o 75 10 -10 / 0.5)', {}, 'color(--din99o 75 10 -10 / 0.5)'),
        # Test alpha
        ('color(--din99o 75 10 -10)', {'alpha': True}, 'color(--din99o 75 10 -10 / 1)'),
        ('color(--din99o 75 10 -10 / 0.5)', {'alpha': False}, 'color(--din99o 75 10 -10)'),
        # Test None
        ('color(--din99o none 10 -10)', {}, 'color(--din99o 0 10 -10)'),
        ('color(--din99o none 10 -10)', {'none': True}, 'color(--din99o none 10 -10)'),
        # Test Fit (not bound)
        ('color(--din99o 120 10 -10)', {}, 'color(--din99o 120 10 -10)'),
        ('color(--din99o 120 10 -10)', {'fit': False}, 'color(--din99o 120 10 -10)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestDIN99oProperties(util.ColorAsserts, unittest.TestCase):
    """Test DIN99o."""

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--din99o 90% 50 -20 / 1)')
        self.assertEqual(c['lightness'], 90)
        c['lightness'] = 80
        self.assertEqual(c['lightness'], 80)

    def test_a(self):
        """Test `a`."""

        c = Color('color(--din99o 90% 50 -20 / 1)')
        self.assertEqual(c['a'], 50)
        c['a'] = 40
        self.assertEqual(c['a'], 40)

    def test_b(self):
        """Test `b`."""

        c = Color('color(--din99o 90% 50 -20 / 1)')
        self.assertEqual(c['b'], -20)
        c['b'] = -10
        self.assertEqual(c['b'], -10)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--din99o 90% 50 -20 / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)



"""Calculate XYZ conversion matrices."""
import sys
import os

sys.path.insert(0, os.getcwd())

from coloraide import util  # noqa: E402
from coloraide import algebra as alg  # noqa: E402

white_d65 = util.xy_to_xyz((0.31270, 0.32900))
white_d50 = util.xy_to_xyz((0.34570, 0.35850))
white_aces = util.xy_to_xyz((0.32168, 0.33767))


def pprint(value):
    """Print the matrix."""
    print('[', end='')
    first = True
    for v in value:
        if first:
            first = False
        else:
            print(',\n ', end='')
        print(v, end='')
    print(']')


def get_matrix(wp, space):
    """Get the matrices for the specified space."""

    if space == 'srgb':
        x = [0.64, 0.30, 0.15]
        y = [0.33, 0.60, 0.06]
    elif space == 'display-p3':
        x = [0.68, 0.265, 0.150]
        y = [0.32, 0.69, 0.060]
    elif space == 'rec2020':
        x = [0.708, 0.17, 0.131]
        y = [0.292, 0.797, 0.046]
    elif space == 'a98-rgb':
        x = [0.64, 0.21, 0.15]
        y = [0.33, 0.71, 0.06]
    elif space == 'prophoto-rgb':
        x = [0.7347, 0.1596, 0.0366]
        y = [0.2653, 0.8404, 0.0001]
    elif space == 'aces-ap0':
        x = [0.7347, 0.0, 0.0001]
        y = [0.2653, 1.0, -0.0770]
    elif space == 'aces-ap1':
        x = [0.713, 0.165, 0.128]
        y = [0.293, 0.830, 0.044]
    else:
        raise ValueError

    m = alg.transpose([util.xy_to_xyz(xy) for xy in zip(x, y)])
    mi = alg.inv(m)
    rgb = alg.dot(mi, wp)
    rgb2xyz = alg.multiply(m, rgb)
    xyz2rgb = alg.inv(rgb2xyz)

    return rgb2xyz, xyz2rgb


if __name__ == "__main__":
    print('===== sRGB =====')
    to_xyz, from_xyz = get_matrix(white_d65, 'srgb')
    print('--- rgb -> xyz ---')
    pprint(to_xyz)
    print('--- xyz -> rgb ---')
    pprint(from_xyz)

    print('===== Display P3 =====')
    to_xyz, from_xyz = get_matrix(white_d65, 'display-p3')
    print('--- rgb -> xyz ---')
    pprint(to_xyz)
    print('--- xyz -> rgb ---')
    pprint(from_xyz)

    print('===== Adobe 98 =====')
    to_xyz, from_xyz = get_matrix(white_d65, 'a98-rgb')
    print('--- rgb -> xyz ---')
    pprint(to_xyz)
    print('--- xyz -> rgb ---')
    pprint(from_xyz)

    print('===== Rec.2020 =====')
    to_xyz, from_xyz = get_matrix(white_d65, 'rec2020')
    print('--- rgb -> xyz ---')
    pprint(to_xyz)
    print('--- xyz -> rgb ---')
    pprint(from_xyz)

    print('===== ProPhoto =====')
    to_xyz, from_xyz = get_matrix(white_d50, 'prophoto-rgb')
    print('--- rgb -> xyz ---')
    pprint(to_xyz)
    print('--- xyz -> rgb ---')
    pprint(from_xyz)

    print('===== ACES =====')
    to_xyz, from_xyz = get_matrix(white_aces, 'aces-ap0')
    print('--- rgb -> xyz ---')
    pprint(to_xyz)
    print('--- xyz -> rgb ---')
    pprint(from_xyz)

    print('===== ACEScg =====')
    to_xyz, from_xyz = get_matrix(white_aces, 'aces-ap1')
    print('--- rgb -> xyz ---')
    pprint(to_xyz)
    print('--- xyz -> rgb ---')
    pprint(from_xyz)



"""Uncalibrated, naive CMY color space."""
from __future__ import annotations
from ..spaces import Space
from ..channels import Channel
from ..cat import WHITES
from ..types import Vector
from .. import algebra as alg
import math


def srgb_to_cmy(rgb: Vector) -> Vector:
    """Convert sRGB to CMY."""

    return [1 - c for c in rgb]


def cmy_to_srgb(cmy: Vector) -> Vector:
    """Convert CMY to sRGB."""

    return [1 - c for c in cmy]


class CMY(Space):
    """The CMY color class."""

    BASE = "srgb"
    NAME = "cmy"
    SERIALIZE = ("--cmy",)  # type: tuple[str, ...]
    CHANNELS = (
        Channel("c", 0.0, 1.0, bound=True),
        Channel("m", 0.0, 1.0, bound=True),
        Channel("y", 0.0, 1.0, bound=True)
    )
    CHANNEL_ALIASES = {
        "cyan": 'c',
        "magenta": 'm',
        "yellow": 'y'
    }
    WHITE = WHITES['2deg']['D65']

    def is_achromatic(self, coords: Vector) -> bool:
        """Test if color is achromatic."""

        black = [1, 1, 1]
        for x in alg.vcross(coords, black):
            if not math.isclose(0.0, x, abs_tol=1e-4):
                return False
        return True

    def to_base(self, coords: Vector) -> Vector:
        """To sRGB."""

        return cmy_to_srgb(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From sRGB."""

        return srgb_to_cmy(coords)



"""LCh D65 class."""
from __future__ import annotations
from ..cat import WHITES
from .lch import LCh
from ..channels import Channel, FLG_ANGLE


class LChD65(LCh):
    """LCh D65 class."""

    BASE = "lab-d65"
    NAME = "lch-d65"
    SERIALIZE = ("--lch-d65",)
    WHITE = WHITES['2deg']['D65']
    CHANNELS = (
        Channel("l", 0.0, 100.0),
        Channel("c", 0.0, 160.0, limit=(0.0, None)),
        Channel("h", 0.0, 360.0, flags=FLG_ANGLE)
    )



"""Test HCT."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color, NaN
import pytest


class TestHCT(util.ColorAssertsPyTest):
    """Test HCT."""

    COLORS = [
        ('red', 'color(--hct 27.41 113.36 53.237)'),
        ('orange', 'color(--hct 71.257 60.528 74.934)'),
        ('yellow', 'color(--hct 111.05 75.504 97.139)'),
        ('green', 'color(--hct 142.23 71.136 46.228)'),
        ('blue', 'color(--hct 282.76 87.228 32.301)'),
        ('indigo', 'color(--hct 310.96 60.765 20.47)'),
        ('violet', 'color(--hct 331.49 65.001 69.695)'),
        ('white', 'color(--hct 209.54 2.8716 100)'),
        ('gray', 'color(--hct 209.54 1.8977 53.585)'),
        ('black', 'color(--hct 209.55 0 0)'),
        # Test color
        ('color(--hct 270 30 100)', 'color(--hct 270 30 100)'),
        ('color(--hct 270 30 100 / 0.5)', 'color(--hct 270 30 100 / 0.5)'),
        ('color(--hct 50% 50% 50% / 50%)', 'color(--hct 180 72.5 50 / 0.5)'),
        ('color(--hct none none none / none)', 'color(--hct none none none / none)'),
        # Test percent ranges
        ('color(--hct 0% 0% 0%)', 'color(--hct 0 0 none)'),
        ('color(--hct 100% 100% 100%)', 'color(--hct 360 145 100 / 1)'),
        ('color(--hct -100% -100% -100%)', 'color(--hct -360 0 0 / 1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('hct'), Color(color2))


class TestHCTSerialize(util.ColorAssertsPyTest):
    """Test HCT serialization."""

    COLORS = [
        # Test color
        ('color(--hct 50 30 75 / 0.5)', {}, 'color(--hct 50 30 75 / 0.5)'),
        # Test alpha
        ('color(--hct 50 30 75)', {'alpha': True}, 'color(--hct 50 30 75 / 1)'),
        ('color(--hct 50 30 75 / 0.5)', {'alpha': False}, 'color(--hct 50 30 75)'),
        # Test None
        ('color(--hct 50 30 none)', {}, 'color(--hct 50 30 0)'),
        ('color(--hct 50 30 none)', {'none': True}, 'color(--hct 50 30 none)'),
        # Test Fit (not bound)
        ('color(--hct 50 230 75)', {}, 'color(--hct 50 230 75)'),
        ('color(--hct 50 230 75)', {'fit': False}, 'color(--hct 50 230 75)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestHCTMisc(util.ColorAsserts, unittest.TestCase):
    """Test miscellaneous cases."""

    def test_wide_gamut(self):
        """Test wide gamut conversion logic."""

        c1 = Color('hct', (330.2920013462662, 484.58580542051413, 91.57114928085913))
        c2 = c1.convert('xyy').convert('hct')
        self.assertColorEqual(c1, c2)


class TestHCTPoperties(util.ColorAsserts, unittest.TestCase):
    """Test HCT properties."""

    def test_names(self):
        """Test LCh-ish names."""

        self.assertEqual(Color('color(--hct 111.05 75.504 97.139)')._space.names(), ('t', 'c', 'h'))

    def test_h(self):
        """Test `h`."""

        c = Color('color(--hct 111.05 75.504 97.139)')
        self.assertEqual(c['h'], 111.05)
        c['h'] = 270
        self.assertEqual(c['h'], 270)

    def test_c(self):
        """Test `c`."""

        c = Color('color(--hct 111.05 75.504 97.139)')
        self.assertEqual(c['c'], 75.504)
        c['c'] = 30
        self.assertEqual(c['c'], 30)

    def test_t(self):
        """Test `t`."""

        c = Color('color(--hct 111.05 75.504 97.139)')
        self.assertEqual(c['t'], 97.139)
        c['t'] = 50
        self.assertEqual(c['t'], 50)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--hct 111.05 75.504 97.139)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('hct', [NaN, 20, 30], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_none_input(self):
        """Test `none` null."""

        c = Color('color(--hct none 20 30 / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_null_normalization_min_chroma(self):
        """Test minimum saturation."""

        c = Color(Color('white').convert('hct').to_string()).normalize()
        self.assertTrue(c.is_nan('hue'))

        c = Color(Color('gray').convert('hct').to_string()).normalize()
        self.assertTrue(c.is_nan('hue'))

        c = Color(Color('darkgray').convert('hct').to_string()).normalize()
        self.assertTrue(c.is_nan('hue'))

    def test_achromatic_hue(self):
        """Test that all RGB-ish colors convert to HCT with a null hue."""

        for space in ('srgb', 'display-p3', 'rec2020', 'a98-rgb', 'prophoto-rgb'):
            for x in range(0, 256):
                color = Color('color({space} {num:f} {num:f} {num:f})'.format(space=space, num=x / 255))
                color2 = color.convert('hct')
                self.assertTrue(color2.is_nan('hue'))


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('#222222').convert('hct').is_achromatic(), True)
        self.assertEqual(
            Color('srgb', [0.000000001] * 3).convert('hct').set('c', lambda x: x + 1e-8).is_achromatic(),
            True
        )
        self.assertEqual(Color('hct', [270, 0.00001, NaN]).is_achromatic(), True)
        self.assertEqual(Color('hct', [270, NaN, 0]).is_achromatic(), True)
        self.assertEqual(Color('hct', [270, 50, 0]).is_achromatic(), True)
        self.assertEqual(Color('hct', [270, 50, NaN]).is_achromatic(), True)
        self.assertEqual(Color('hct', [270, NaN, 20]).is_achromatic(), False)
        self.assertEqual(Color('hct', [270, NaN, NaN]).is_achromatic(), True)



"""HCT gamut mapping."""
from __future__ import annotations
from ..gamut.fit_lch_chroma import LChChroma


class HCTChroma(LChChroma):
    """HCT chroma gamut mapping class."""

    NAME = "hct-chroma"

    EPSILON = 0.001
    LIMIT = 0.02
    DE = "hct"
    SPACE = "hct"
    MIN_LIGHTNESS = 0
    MAX_LIGHTNESS = 100
    MIN_CONVERGENCE = 0.0001



"""
Calculate `oklab` matrices.

Björn Ottosson, in his original calculations, used a different white point than
what CSS and most other people use. At the CSS repository, he commented on
how to calculate the M1 matrix using the exact same white point as CSS. He
provided the initial matrix which we call M0.
https://github.com/w3c/csswg-drafts/issues/6642#issuecomment-945714988.
This M0 matrix is used to create a precise matrix to convert XYZ to using
the D65 white point as specified by CSS and used by most people. We use
the D65 chromaticity coordinates of `(0.31270, 0.32900)` which is documented
and used for sRGB as the standard. There are likely implementations unaware
that the should, or even how to adapt the Oklab M1 matrix to their white point
as this is not documented in the author's Oklab blog post, but is buried in a
CSS repository issue.

Additionally, the documented M2 matrix is specified as 32 bit values, and the
inverse is calculated directly from the this 32 bit matrix. The forward and
reverse transform is calculated to perfect convert 32 bit values, but when
translating to 64 bit, especially for achromatic colors, adds a lot of noise
after about 7 - 8 digits, the precision of 32 bit floats.

To provide an M2 matrix that works better for 64 bit, we take the inverse M2,
which provides a perfect transforms to white from Oklab `[1, 0, 0]` in 32 bit
floating point. We process matrix as a float 32 bit values and emit it as a 64
double values, ~17 digit double accuracy. Then we apply a slight correction to
account for the 64 bit noise to ensure it converts a perfect Oklab `[1, 0, 0]`
to LMS `[1, 1, 1]`. Once corrected, we then calculate the forward matrix. This
gives us a transform in 64 bit that drives chroma extremely close to zero for
64 bit doubles and maintains 32 bit precision of up to about 7 digits, the 32
bit accuracy limit (~7.22).

To demonstrate that our 64 bit converted matrices work as we claim and does not
alter the intent of the values, we can observe by comparing the documented matrices
(adjusting for our white point).

Below we demonstrate by first using the documented 32 bit M2 matrix (adjusting the
M1 for our white point). This is what most implementations do, though some may not
properly correct the M1 matrix for their white point. Notice how the lightness for
white is only accurate up to about 7 digits making the expected value of 1 not very
accurate. Also notice that a and b do not resolve as close to 0. The a value is
pretty good, but the b value is substantially worse. Also notice the first 7 digits
(the 32 bit precision) for red, green, and blue as they will be used for comparison.

```
>>> from coloraide.everything import ColorAll as Color
>>> import numpy as np
>>> Color('white').convert('oklab')[:]
[0.9999999935000001, -1.6653345369377348e-16, 3.729999997759137e-08, 1.0]
>>> [np.float32(c) for c in Color('red').convert('oklab', norm=False)[:]]
[0.6279554, 0.22486307, 0.1258463, 1.0]
>>> [np.float32(c) for c in Color('green').convert('oklab', norm=False)[:]]
[0.51975185, -0.14030233, 0.107675895, 1.0]
>>> [np.float32(c) for c in Color('blue').convert('oklab', norm=False)[:]]
[0.4520137, -0.03245697, -0.31152815, 1.0]
```

When we use our 64 bit adjusted M2 matrix, we now get a precise 1 for lightness
when converting white and get zero or nearly zero for a and b. When comparing the
first 7 digits to the previous example we get the same values. Anything after
7 digits is not guaranteed to be the same.

```
>>> from coloraide.everything import ColorAll as Color
>>> import numpy as np
>>> Color('white').convert('oklab')[:]
[1.0, -5.551115123125783e-17, 0.0, 1.0]
>>> [np.float32(c) for c in Color('red').convert('oklab', norm=False)[:]]
[0.6279554, 0.22486307, 0.12584628, 1.0]
>>> [np.float32(c) for c in Color('green').convert('oklab', norm=False)[:]]
[0.51975185, -0.14030233, 0.10767588, 1.0]
>>> [np.float32(c) for c in Color('blue').convert('oklab', norm=False)[:]]
[0.45201373, -0.032456975, -0.31152818, 1.0]
```

Okhsl is completely calculated using 32 bit floats as that is how the author
provided the algorithm, but we can see that when we calculate the new coefficients,
using our M1 and 64 bit adjusted M2 matrices, that we preserve the 32 precision.
Anything after ~7 digits is just noise due to the differences in 32 bit and 64 bit.

Comparing to the actual values returned using the author's code in his Okhsl and Okhsv
color pickers:

```
// Okhsl
> var value = srgb_to_okhsl(255, 255, 255); value[0] *= 360; value
[89.87556309590242, 0.5582831888483675, 0.9999999923961898]
> var value = srgb_to_okhsl(255, 0, 0); value[0] *= 360; value
[29.23388519234263, 1.0000000001433997, 0.5680846525040862]
> var value = srgb_to_okhsl(0, 255, 0); value[0] *= 360; value
[142.49533888780996, 0.9999999700728788, 0.8445289645307816]
> var value = srgb_to_okhsl(0, 0, 255); value[0] *= 360; value
[264.052020638055, 0.9999999948631134, 0.3665653394260194]

// Okhsv
> var value = srgb_to_okhsv(255, 255, 255); value[0] *= 360; value
[89.87556309590242, 1.0347523928230576e-7, 1.000000027003774]
> var value = srgb_to_okhsv(255, 0, 0); value[0] *= 360; value
[29.23388519234263, 0.9995219692256989, 1.0000000001685625]
> var value = srgb_to_okhsv(0, 255, 0); value[0] *= 360; value
[142.49533888780996, 0.9999997210415695, 0.9999999884428648]
> var value = srgb_to_okhsv(0, 0, 255); value[0] *= 360; value
[264.052020638055, 0.9999910912349018, 0.9999999646150918]
```

And then ours. Ignoring the authors hue and our hue results for white
and the oddly high chroma for the author's achromatic white in Okhsl
(both of which are meaningless in an achromatic color), we can see that
that we match quite well up to ~7 digits.

```
# Okhsl
>>> Color('white').convert('okhsl', norm=False)[:]
[180.0, 0.0, 1.0, 1.0]
>>> Color('#ff0000').convert('okhsl', norm=False)[:]
[29.233880279627876, 1.0000001765854427, 0.5680846563197033, 1.0]
>>> Color('#00ff00').convert('okhsl', norm=False)[:]
[142.4953450414438, 1.0000000000000009, 0.8445289714936317, 1.0]
>>> [264.05202261637004, 1.0000000005848086, 0.36656533918708145, 1.0]
[264.05202261637004, 1.0000000005848086, 0.36656533918708145, 1.0]
# Okhsv
>>> Color('white').convert('okhsv', norm=False)[:]
[180.0, 0.0, 1.0, 1.0]
>>> Color('#ff0000').convert('okhsv', norm=False)[:]
[29.233880279627876, 1.0000004019360378, 0.9999999999999994, 1.0]
>>> Color('#00ff00').convert('okhsv', norm=False)[:]
[142.4953450414438, 0.9999998662471965, 1.0000000000000004, 1.0]
>>> Color('#0000ff').convert('okhsv', norm=False)[:]
[264.05202261637004, 1.000000002300706, 0.9999999999999999, 1.0]
"""
import sys
import os
import struct

sys.path.insert(0, os.getcwd())

import tools.calc_xyz_transform as xyzt  # noqa: E402
from coloraide import algebra as alg  # noqa: E402


def pprint(value):
    """Print the matrix."""
    print('[', end='')
    first = True
    for v in value:
        if first:
            first = False
        else:
            print(',\n ', end='')
        print(v, end='')
    print(']')


float32 = alg.vectorize(lambda value: struct.unpack('f', struct.pack('f', value))[0])


# Calculated using our own `calc_xyz_transform.py`
RGB_TO_XYZ, XYZ_TO_RGB = xyzt.get_matrix(xyzt.white_d65, 'srgb')

# Matrix provided by the author of Oklab to allow for calculating a precise M1 matrix
# using any white point.
M0 = [
    [0.77849780, 0.34399940, -0.12249720],
    [0.03303601, 0.93076195, 0.03620204],
    [0.05092917, 0.27933344, 0.66973739]
]

# Calculate XYZ to LMS and LMS to XYZ using our white point.
XYZ_TO_LMS = alg.divide(M0, alg.outer(alg.dot(M0, xyzt.white_d65), alg.ones(3)))

# Calculate the inverse
LMS_TO_XYZ = alg.inv(XYZ_TO_LMS)

# Calculate linear sRGB to LMS (used for Okhsl and Okhsv)
SRGBL_TO_LMS = alg.dot(XYZ_TO_LMS, RGB_TO_XYZ)
LMS_TO_SRGBL = alg.inv(SRGBL_TO_LMS)

# Oklab specifies the following matrix as M1 along with the inverse.
# ```
# LMS3_TO_OKLAB = [
#     [0.2104542553, 0.7936177850, -0.0040720468],
#     [1.9779984951, -2.4285922050, 0.4505937099],
#     [0.0259040371, 0.7827717662, -0.8086757660]
# ]
# ```
# But since the matrix is provided in 32 bit, we are not able to get the
# inverse, and in return, we do not get a good resolution to `[1, 0, 0]`
# for white. In order to adjust for this, take documented 32 bit inverse
# matrix # that expects `[1, 0, 0]` and work backwards to calculate the 64
# bit version. Make sure to process the value as 32 bit but emit it as 64
# bit, then correct the 64 bit matrix to ensure it still aligns for `[1, 0, 0]`.
OKLAB_TO_LMS3 = float32(
    [
        [1.0, 0.3963377774, 0.2158037573],
        [1.0, -0.1055613458, -0.0638541728],
        [1.0, -0.0894841775, -1.2914855480]
    ]
)

# Calculate what we expect the ideal translation for D65 white to be.
correct = alg.diag([alg.nth_root(c, 3) for c in alg.dot(XYZ_TO_LMS, xyzt.white_d65)])

# Adjust to target a precise translation for white
lms3 = alg.diag(alg.dot(OKLAB_TO_LMS3, [1.0, 0.0, 0.0]))
OKLAB_TO_LMS3 = alg.multi_dot([OKLAB_TO_LMS3, lms3, alg.inv(correct)])

# Calculate the inverse
LMS3_TO_OKLAB = alg.inv(OKLAB_TO_LMS3)

if __name__ == "__main__":
    print('===== sRGB Linear -> lms =====')
    pprint(SRGBL_TO_LMS)
    print('===== lms -> sRGB Linear =====')
    pprint(LMS_TO_SRGBL)
    print('===== XYZ D65 Linear -> lms =====')
    pprint(XYZ_TO_LMS)
    print('===== lms -> XYZ D65 =====')
    pprint(LMS_TO_XYZ)
    print('===== lms ** 1/3 -> Oklab =====')
    pprint(LMS3_TO_OKLAB)
    print('===== Oklab -> lms ** 1/3 =====')
    pprint(OKLAB_TO_LMS3)



"""Contrast."""
from __future__ import annotations
from abc import ABCMeta, abstractmethod
from ..types import Plugin
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class ColorContrast(Plugin, metaclass=ABCMeta):
    """Color contrast plugin class."""

    NAME = ''

    @abstractmethod
    def contrast(self, color1: Color, color2: Color, **kwargs: Any) -> float:
        """Get the contrast of the two provided colors."""


def contrast(name: str | None, color1: Color, color2: Color, **kwargs: Any) -> float:
    """Get the appropriate contrast plugin."""

    if name is None:
        name = color1.CONTRAST

    method = color1.CONTRAST_MAP.get(name)
    if not method:
        raise ValueError("'{}' contrast method is not supported".format(name))

    return method.contrast(color1, color2, **kwargs)



"""Linear Display-p3 color class."""
from __future__ import annotations
from ..cat import WHITES
from .srgb import sRGB
from .. import algebra as alg
from ..types import Vector

RGB_TO_XYZ = [
    [0.4865709486482161, 0.26566769316909306, 0.1982172852343625],
    [0.22897456406974875, 0.6917385218365063, 0.079286914093745],
    [-3.972075516933487e-17, 0.04511338185890263, 1.043944368900976]
]

XYZ_TO_RGB = [
    [2.4934969119414254, -0.9313836179191239, -0.40271078445071684],
    [-0.8294889695615747, 1.7626640603183465, 0.02362468584194358],
    [0.03584583024378446, -0.0761723892680418, 0.9568845240076872]
]


def lin_p3_to_xyz(rgb: Vector) -> Vector:
    """
    Convert an array of linear-light image-p3 values to CIE XYZ using  D65 (no chromatic adaptation).

    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    """

    # 0 was computed as -3.972075516933488e-17
    return alg.dot(RGB_TO_XYZ, rgb, dims=alg.D2_D1)


def xyz_to_lin_p3(xyz: Vector) -> Vector:
    """Convert XYZ to linear-light P3."""

    return alg.dot(XYZ_TO_RGB, xyz, dims=alg.D2_D1)


class DisplayP3Linear(sRGB):
    """Linear Display-p3 class."""

    BASE = "xyz-d65"
    NAME = "display-p3-linear"
    SERIALIZE = ('--display-p3-linear',)
    WHITE = WHITES['2deg']['D65']

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from Linear Display P3."""

        return lin_p3_to_xyz(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to Linear Display P3."""

        return xyz_to_lin_p3(coords)



"""Gamut handling."""
from __future__ import annotations
from .. import algebra as alg
from ..channels import FLG_ANGLE
from abc import ABCMeta, abstractmethod
from ..types import Plugin
from typing import TYPE_CHECKING, Any
from .. import util
from . import pointer

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color

__all__ = ('clip_channels', 'verify', 'Fit', 'pointer')


def clip_channels(color: Color, nans: bool = True) -> None:
    """Clip channels."""

    for i, value in enumerate(color[:-1]):

        chan = color._space.CHANNELS[i]

        # Wrap the angle. Not technically out of gamut, but we will clean it up.
        if chan.flags & FLG_ANGLE:
            color[i] = util.constrain_hue(value)

        # Ignore undefined or unbounded channels
        if not chan.bound or alg.is_nan(value):
            continue

        # Fit value in bounds.
        color[i] = alg.clamp(value, chan.low, chan.high)


def verify(color: Color, tolerance: float) -> bool:
    """Verify the values are in bound."""

    for i, value in enumerate(color[:-1]):
        chan = color._space.CHANNELS[i]

        # Ignore undefined channels, angles which wrap, and unbounded channels
        if chan.flags & FLG_ANGLE or not chan.bound or alg.is_nan(value):
            continue

        a = chan.low
        b = chan.high

        # Check if bounded values are in bounds
        if (a is not None and value < (a - tolerance)) or (b is not None and value > (b + tolerance)):
            return False
    return True


class Fit(Plugin, metaclass=ABCMeta):
    """Fit plugin class."""

    NAME = ''

    @abstractmethod
    def fit(self, color: Color, **kwargs: Any) -> None:
        """Get coordinates of the new gamut mapped color."""



"""
Render 3D plots of sRGB in other color spaces.

In order to render things fast and allow for reasonable
performance, we render the outer shell of the space.

Only sRGB and related sRGB cylindrical models are precise,
all others are approximations of the shape due to our
approach.
"""
import matplotlib.pyplot as plt
import sys
import os
import argparse
import numpy as np

sys.path.insert(0, os.getcwd())

try:
    from coloraide_extras.everything import ColorAll as Color
except ImportError:
    from coloraide.everything import ColorAll as Color
from coloraide.spaces import Cylindrical, LChish, Labish, HSLish, HSVish  # noqa: E402
from coloraide import algebra as alg  # noqa: E402

# Special cases for certain color spaces
color_options = {
    'hpluv': {'gamut_self': True, 'factor': 100},
    'hwb': {'gamut_self': True, 'zsort': 'max'}
}


def create_custom_hsl(gamut):
    """Create a custom color object that has access to special `hsl-gamut` space to map surface in."""

    cs = Color.CS_MAP[gamut]

    class HSL(type(Color.CS_MAP['hsl'])):
        NAME = 'hsl-{}'.format(gamut)
        BASE = gamut
        GAMUT_CHECK = gamut
        WHITE = cs.WHITE
        DYAMIC_RANGE = cs.DYNAMIC_RANGE

    class ColorCyl(Color):
        """Custom color."""

    ColorCyl.register(HSL())

    return ColorCyl


def cyl_disc(ColorCyl, space, gamut, factor, location, mode, x, y, z, cmap, resolution, opacity):
    """Plot cylindrical disc."""

    cs = ColorCyl.CS_MAP[space]
    if isinstance(cs, (HSVish, HSLish)):
        space_type = 'hslish/hsvish'
    else:
        space_type = space

    b = 0.0 if location == 'bottom' else 1.0 * factor
    start, end = (1.0 * factor, 0.0) if location == 'bottom' else (0.0, 1.0 * factor)

    # Simple style, just create a dot in the center to complete the mesh
    if mode == 'simple':
        c = ColorCyl(gamut, [0, 0, b], opacity).convert(space, norm=False)
        if space_type == 'hslish/hsvish':
            hue, saturation, lightness = c._space.indexes()
            a, b = alg.polar_to_rect(c[saturation], c[hue])
            x.append([a] * resolution)
            y.append([b] * resolution)
            z.append([c[lightness]] * resolution)
        else:
            hue = c._space.hue_index()
            a, b = alg.polar_to_rect(c[1], c[hue])
            x.append([a] * resolution)
            y.append([b] * resolution)
            z.append([c[2]] * resolution)
        s = c.convert('srgb')
        if not s.in_gamut():
            s.fit()
        cmap.append([s.to_string(hex=True)] * resolution)
        return

    # Create concentric circles
    c1 = ColorCyl(gamut, [0, start, b], opacity)
    c2 = ColorCyl(gamut, [359.99999, start, b], opacity)
    s1 = ColorCyl.steps([c1, c2], steps=resolution, space=gamut, hue='specified')
    s2 = [c.clone().set('w', end) for c in s1]

    step = int(resolution / 2)
    for r in range(step + 1):
        if location == 'bottom':
            r = step - r
        x.append([])
        y.append([])
        z.append([])
        cmap.append([])
        for t1, t2 in zip(s1, s2):
            c = t1.mix(t2, r / step, space=gamut, hue='specified')
            c.convert(space, norm=False, in_place=True)
            if space_type == 'hslish/hsvish':
                hue, saturation, lightness = c._space.indexes()
                a, b = alg.polar_to_rect(c[saturation], c[hue])
                x[-1].append(a)
                y[-1].append(b)
                z[-1].append(c[lightness])
            else:
                hue = c._space.hue_index()
                a, b = alg.polar_to_rect(c[1], c[hue])
                x[-1].append(a)
                y[-1].append(b)
                z[-1].append(c[2])
            s = c.convert('srgb')
            if not s.in_gamut():
                s.fit()
            cmap[-1].append(s.to_string(hex=True))


def render_space(space, gamut, ax, mode, resolution, opacity):
    """
    Render the space with the given resolution and factor.

    Any normal RGB gamut can be used, assuming it has ranges [0,1].
    The HSL space is used to map as it makes it easy to get the outermost
    colors of RGB spaces.

    There are a number of special options to handle RGB cylinder spaces.
    Some require us to generate top and bottom caps. This is mainly because
    these models are special and don't quite fit generically with things like
    Lab, LCh, and rectangular spaces.
    """

    x = []
    y = []
    z = []
    cmap = []

    cs = Color.CS_MAP[space]

    # What kind of space are we dealing with?
    is_cyl = isinstance(cs, Cylindrical)
    if isinstance(cs, Labish):
        space_type = 'labish'
    elif isinstance(cs, LChish):
        space_type = 'lchish'
    elif isinstance(cs, HSVish):
        space_type = 'hsvish'
    elif isinstance(cs, HSLish):
        space_type = 'hslish'
    else:
        space_type = space

    # Create an HSL color space for the gamut for easily creating
    # a shell of the gamut. If we are rendering in sRGB gamut, we
    # can use the already present HSL space. Some spaces are special
    # and we must use themselves as the gamut, skip this step if
    # required.
    if color_options.get(space, {}).get('gamut_self'):
        ColorCyl = Color
        gamut_space = space
    elif gamut != 'srgb':
        ColorCyl = create_custom_hsl(gamut)
        gamut_space = 'hsl-{}'.format(gamut)
    else:
        ColorCyl = Color
        gamut_space = 'hsl'

    # Does the space require special factoring?
    factor = color_options.get(space, {}).get('factor', 1)

    # Calculate top disc
    if space_type in ('hslish', 'hwb'):
        cyl_disc(
            ColorCyl,
            space,
            gamut_space,
            factor,
            'top',
            'full' if space_type == 'hwb' else 'simple',
            x,
            y,
            z,
            cmap,
            resolution,
            opacity
        )

    # Calculate top cylinder
    c1 = ColorCyl(gamut_space, [0, 1 * factor, 1 * factor], opacity)
    c2 = ColorCyl(gamut_space, [0, 1 * factor, 0], opacity)
    c3 = ColorCyl(gamut_space, [359.99999, 1 * factor, 1 * factor], opacity)
    c4 = ColorCyl(gamut_space, [359.99999, 1 * factor, 0], opacity)
    s1 = ColorCyl.steps([c1, c2], steps=int(resolution / 2), space=gamut_space)
    s2 = ColorCyl.steps([c3, c4], steps=int(resolution / 2), space=gamut_space)

    for t1, t2 in zip(s1, s2):
        x.append([])
        y.append([])
        z.append([])
        cmap.append([])
        for c in ColorCyl.steps([t1, t2], steps=resolution, space=gamut_space, hue='specified'):
            c = c.convert(space, norm=False)

            if space_type in ('hslish', 'hsvish'):
                hue, saturation, lightness = c._space.indexes()
                a, b = alg.polar_to_rect(c[saturation], c[hue])
                x[-1].append(a)
                y[-1].append(b)
                z[-1].append(c[lightness])
            elif space_type == 'lchish':
                lightness, chroma, hue = c._space.indexes()
                a, b = alg.polar_to_rect(c[chroma], c[hue])
                x[-1].append(a)
                y[-1].append(b)
                z[-1].append(c[lightness])
            elif space_type == 'labish':
                lightness, a, b = c._space.indexes()
                x[-1].append(c[a])
                y[-1].append(c[b])
                z[-1].append(c[lightness])
            elif is_cyl:
                hue = c._space.hue_index()
                if hue == 0:
                    i, j = 1, 2
                elif hue == 1:
                    i, j = 0, 2
                else:
                    i, j = 0, 1
                a, b = alg.polar_to_rect(c[i], c[hue])
                x[-1].append(a)
                y[-1].append(b)
                z[-1].append(c[j])
            else:
                x[-1].append(c[0])
                y[-1].append(c[1])
                z[-1].append(c[2])

            s = c.convert('srgb')
            if not s.in_gamut():
                s.fit()
            cmap[-1].append(s.to_string(hex=True))

    # Calculate bottom disc
    if space_type in ('hslish', 'hsvish', 'hwb'):
        cyl_disc(
            ColorCyl,
            space,
            gamut_space,
            factor,
            'bottom',
            'full' if space_type == 'hwb' else 'simple',
            x,
            y,
            z,
            cmap,
            resolution,
            opacity
        )

    # Plot the data
    if mode == 'wireframe':
        ax.plot_wireframe(
            np.asarray(x), np.asarray(y), np.asarray(z),
            color=Color('blue').set('alpha', opacity).to_string(hex=True),
            rcount=resolution,
            ccount=resolution,
            antialiased=True
        )
    elif mode == 'scatter':
        scatter_cmap = []
        for m in cmap:
            scatter_cmap.extend(m)
        ax.scatter(
            np.asarray(x).ravel(), np.asarray(y).ravel(), np.asarray(z).ravel(),
            c=scatter_cmap,
            s=20 * 4
        )
    else:
        # Plot the data as a 3D surface.
        ax.plot_surface(
            np.asarray(x), np.asarray(y), np.asarray(z),
            rcount=resolution,
            ccount=resolution,
            zsort=color_options.get(space, {}).get('zsort', 'average'),
            facecolors=cmap,
            shade=False,
            antialiased=True,
        )


def plot_gamut_in_space(
    space,
    gamut,
    title="",
    dark=False,
    rotate_elev=30.0,
    rotate_azim=-60.0,
    mode='surface',
    resolution=200,
    opacity=1.0
):
    """Plot the given space in sRGB."""

    # Get names for
    target = Color.CS_MAP[space]
    names = target.CHANNELS
    is_cyl = isinstance(target, Cylindrical)
    is_labish = isinstance(target, Labish)
    is_lchish = isinstance(target, LChish)
    is_hslish_hsvish = isinstance(target, (HSLish, HSVish))

    # Some spaces need us to rearrange the order of the data
    if is_labish:
        c1, c2, c3 = target.indexes()
        axm = [c2, c3, c1]
    elif is_lchish:
        c1, c2, c3 = target.indexes()
        axm = [c3, c2, c1]
    elif is_hslish_hsvish:
        axm = [0, 1, 2]
    else:
        axm = [0, 1, 2]

    # Select the right theme
    if dark:
        plt.style.use('dark_background')
    else:
        plt.style.use('seaborn-v0_8-bright')

    # Setup figure and axis
    figure = plt.figure()
    plt.tight_layout()
    ax = plt.axes(
        projection='3d',
        xlabel=names[axm[0]] if not is_cyl else "{} (0˚ - 360˚)".format(names[axm[0]]),
        ylabel=names[axm[1]],
        zlabel=names[axm[2]]
    )

    # Turn off ticks for cylindrical hue
    if is_cyl:
        ax.xaxis.set_ticks([])
    figure.add_axes(ax)

    # Add title
    acutal_gamut = space if color_options.get(space, {}).get('gamut_self') else gamut
    plt.title(title if title else "'{}' Rendered in '{}'".format(acutal_gamut, space), pad=20)

    # Setup the aspect ratio
    ax.set_box_aspect((1, 1, 1))

    # Render the space
    render_space(space, gamut, ax, mode, resolution, opacity)

    ax.view_init(rotate_elev, rotate_azim)


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='3d_diagrams', description='Plot 3D gamut in a different color spaces.')
    parser.add_argument('--space', '-s', help='Desired space.')
    parser.add_argument(
        '--gamut',
        '-g',
        default='srgb',
        help=(
            'Gamut space to render space in. Gamut space must be bounded and must have channels in the range [0, 1].'
            'As only a shell is rendered for the gamut, the target space should be less than or equal to the size of '
            'the target gamut or there will be areas that do not render. Cylindrical spaces based specifically off '
            'an RGB gamut, such as HSL being based on sRGB, will only be done under the related gamut and will ignore '
            'this option.'
        )
    )
    parser.add_argument('--opacity', default=1.0, type=float, help="opacity")
    parser.add_argument('--render-mode', '-m', default="surface", help="Render mode: surface, wireframe, scatter.")
    parser.add_argument(
        '--resolution', '-r',
        default="200",
        help=(
            "How densely to render the figure. Some spaces need higher resolution to flesh out certain areas, "
            "but it comes at the cost of speed."
        )
    )
    parser.add_argument('--rotate-elev', '-e', default=30.0, type=float, help="Rotate x axis by specified angle.")
    parser.add_argument('--rotate-azim', '-a', default=-60.0, type=float, help="Rotate y axis by specified angle.")
    parser.add_argument('--title', '-t', default='', help="Provide a title for the diagram.")
    parser.add_argument('--dark', action="store_true", help="Use dark theme.")
    parser.add_argument('--output', '-o', default='', help='Output file.')
    parser.add_argument('--dpi', default=200, type=int, help="DPI of image.")
    args = parser.parse_args()

    if args.render_mode not in ('surface', 'wireframe', 'scatter'):
        raise TypeError("Render mode '{}' not supported".format(args.render_mode))

    plot_gamut_in_space(
        args.space,
        args.gamut,
        title=args.title,
        dark=args.dark,
        rotate_elev=args.rotate_elev,
        rotate_azim=args.rotate_azim,
        mode=args.render_mode,
        resolution=int(args.resolution),
        opacity=args.opacity
    )

    if args.output:
        plt.savefig(args.output, dpi=args.dpi)
    else:
        plt.gcf().set_dpi(args.dpi)
        plt.show()

    plt.close()


if __name__ == "__main__":
    sys.exit(main())



"""Distance and Delta E."""
from __future__ import annotations
from abc import ABCMeta, abstractmethod
import math
from .. import algebra as alg
from ..types import ColorInput, Plugin
from typing import TYPE_CHECKING, Any, Sequence

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


def closest(color: Color, colors: Sequence[ColorInput], method: str | None = None, **kwargs: Any) -> Color:
    """Get the closest color."""

    if method is None:
        method = color.DELTA_E

    algorithm = color.DE_MAP.get(method)
    if not algorithm:
        raise ValueError("'{}' is not currently a supported distancing algorithm.".format(method))

    lowest = alg.INF
    closest = None
    for c in colors:
        color2 = color._handle_color_input(c)
        de = algorithm.distance(color, color2, **kwargs)
        if de < lowest:
            lowest = de
            closest = color2

    if closest is None:
        raise ValueError('No colors to compare')

    return closest


def distance_euclidean(color: Color, sample: Color, space: str = "lab-d65") -> float:
    """
    Euclidean distance.

    https://en.wikipedia.org/wiki/Euclidean_distance
    """

    coords1 = color.convert(space, norm=False).coords(nans=False)
    coords2 = sample.convert(space, norm=False).coords(nans=False)

    return math.sqrt(sum((x - y) ** 2.0 for x, y in zip(coords1, coords2)))


class DeltaE(Plugin, metaclass=ABCMeta):
    """Delta E plugin class."""

    NAME = ''

    @abstractmethod
    def distance(self, color: Color, sample: Color, **kwargs: Any) -> float:
        """Get distance between color and sample."""



"""
Calculate IgPgTg matrices.

https://www.ingentaconnect.com/content/ist/jpi/2020/00000003/00000002/art00002#
"""
import sys
import os

sys.path.insert(0, os.getcwd())

from coloraide import algebra as alg  # noqa: E402

m1 = [
    [2.968, 2.741, -0.649],
    [1.237, 5.969, -0.173],
    [-0.318, 0.387, 2.311]
]

m2 = [
    [0.117, 1.464, 0.130],
    [8.285, -8.361, 21.40],
    [-1.208, 2.412, -36.53]
]


def pprint(value):
    """Print the matrix."""
    print('[', end='')
    first = True
    for v in value:
        if first:
            first = False
        else:
            print(',\n ', end='')
        print(v, end='')
    print(']')


if __name__ == "__main__":
    print('===== XYZ to LMS =====')
    pprint(m1)
    print('===== LMS to XYZ =====')
    pprint(alg.inv(m1))
    print('===== LMS to IgPgTg =====')
    pprint(m2)
    print('===== IgPgTg to LMS =====')
    pprint(alg.inv(m2))



"""
Map our achromatic spline calculation against real achromatic response.

Report the delta between our spline and the real world. Also note the highest chroma climbs.
"""
import sys
import argparse
import os
import matplotlib.pyplot as plt
import math

sys.path.insert(0, os.getcwd())

from coloraide.everything import ColorAll as Color  # noqa: E402
from coloraide.spaces.cam16_jmh import Environment, xyz_d65_to_cam16, Achromatic  # noqa: E402
from coloraide.cat import WHITES  # noqa: E402


def main():
    """Main."""

    parser = argparse.ArgumentParser(
        prog='calc_cam16_ucs_jmh_min_m.py',
        description='Calculate min M for achromatic colors in CAM16 UCS JMh and map current spline against real values.'
    )
    # Flag arguments
    parser.add_argument(
        '--res', '-r', type=int, default=50000, help="Resolution to use when calculating range, default is 50000."
    )
    parser.add_argument(
        '--white-point', '-w', type=str, default='2deg:D65', help="White point 'deg:wp', example: '2deg:D65'."
    )
    parser.add_argument(
        '--adapting-luminance', '-a', type=float, default=64 / math.pi * 0.2, help="Adapting luminance."
    )
    parser.add_argument(
        '--background-luminance', '-b', type=float, default=20, help="Background luminace - default 20 (gray world)."
    )
    parser.add_argument(
        '--surround', '-s', type=str, default='average', help="Surround: 'average', 'dim', 'dark'"
    )
    parser.add_argument(
        '--discounting', '-d', action='store_true', help="Enable discounting."
    )
    parser.add_argument(
        '--spline', '-S', type=str, default='catrom', help="Spline to use for approximation of achromatic line"
    )
    parser.add_argument(
        '--tuning', '-t', type=str, action='append',
        help="Spline tuning parameters: start:end:step:scale (int:int:int:float)"
    )
    parser.add_argument(
        '--dump', action='store_true', help="Dump calculated values."
    )
    args = parser.parse_args()

    return run(
        args.white_point,
        args.adapting_luminance,
        args.background_luminance,
        args.surround,
        args.discounting,
        args.spline,
        args.tuning,
        args.res,
        args.dump
    )


def run(white_point, la, ba, surround, discounting, spline, tuning, res, dump):
    """Run."""

    deg, wp = white_point.split(':')
    tune = [[int(i) if e < 3 else float(i) for e, i in enumerate(x.split(':'))] for x in tuning]

    env = Environment(WHITES[deg][wp], la, ba, surround, discounting)
    convert = xyz_d65_to_cam16

    test = Achromatic(spline=spline, env=env)
    test.calc_achromatic_response(tune, env=env)

    color = Color('srgb', [0, 0, 0])
    points1 = {}
    points2 = {}
    diff_over = 0
    diff_under = 0
    max_m = 0
    first = False

    for i in range(1, res + 1):
        div = res / 5
        v = i / div
        if v < 0.001:
            continue
        color.update('srgb', [v, v, v])
        xyz = color.convert('xyz-d65')
        coords = convert(xyz[:-1], env)[:]
        j, m, h = coords[0], coords[5], coords[2]
        if not first:
            print('Starting J: ', j)
            print('Starting M: ', m)
            first = True

        if m > max_m:
            max_m = m

        if not env.discounting:
            domain = test.scale(j)
            calc = test.spline(domain)
        else:
            calc = (j, 1e-08, 0.0)

        delta = calc[1] - m
        if delta >= 0 and delta > diff_over:
            diff_over = delta
        if delta < 0 and abs(delta) > diff_under:
            diff_under = abs(delta)

        points1[j] = (m, h)
        points2[calc[0]] = (calc[1], calc[2])

    print('Delta Over: ', diff_over)
    print('Delta Under: ', diff_under)
    print('Max M: ', max_m)
    if not env.discounting:
        print('Data Points: ', test.spline.length)

    j1 = []
    j2 = []
    m1 = []
    m2 = []
    h1 = []
    h2 = []
    for j in sorted(points1):
        j1.append(j)
        m1.append(points1[j][0])
        h1.append(points1[j][1])
    for j in sorted(points2):
        j2.append(j)
        m2.append(points2[j][0])
        h2.append(points2[j][1])

    figure = plt.figure()

    ax = plt.axes(
        projection='3d',
        xlabel='M',
        ylabel='H',
        zlabel='J'
    )

    # ax.set_aspect('auto')
    ax.set_title('JMh: Delta = {} - Max M = {}'.format(max(diff_over, diff_under), max_m))
    figure.add_axes(ax)

    # Print the calculated line against the real line
    plt.style.use('seaborn-v0_8-whitegrid')
    plt.plot(m1, h1, j1, '.', color='black')
    plt.plot(m2, h2, j2, '.', color='red', markersize=0.5)
    plt.show()

    if dump:
        print('===== Data =====')
        print(test.dump())

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Test miscellaneous API features."""
import unittest
from coloraide import Color, NaN
from coloraide.everything import ColorAll
from . import util
import math


class TestMisc(util.ColorAsserts, unittest.TestCase):
    """Test miscellaneous API features."""

    def test_max_precision(self):
        """Test max precision."""

        self.assertEqual(
            Color('purple').convert('lab').to_string(precision=-1),
            'lab(29.6915242415228703976026736199855804443359375 56.11166698871137015203203191049396991729736328125 -36.29266541032566095736910938285291194915771484375)'  # noqa:  E501
        )

    def test_repr(self):
        """Test string representation."""

        self.assertEqual(str(Color('red')), 'color(srgb 1 0 0 / 1)')

    def test_repr_percent(self):
        """Test string percent representation."""

        self.assertEqual(str(Color('white').convert('lab')), 'color(--lab 100 0 0 / 1)')

    def test_color_indexing(self):
        """Test color indexing."""

        c1 = Color('red')
        self.assertEqual(len(c1), 4)
        self.assertEqual(c1[0], 1)
        self.assertEqual(c1[:], [1, 0, 0, 1])
        c1[2] = 1
        self.assertEqual(c1[2], 1)
        c2 = Color('blue')
        c1[:] = c2[:]
        self.assertColorEqual(c1, c2)

    def test_none(self):
        """Test none."""

        c = Color('color(srgb 1 none 1)')
        self.assertEqual(c[:-1:2], [1, 1])
        self.assertTrue(c.is_nan('green'))
        self.assertTrue(Color('color(srgb 1 1 1 / none)').is_nan('alpha'))

    def test_percent_none(self):
        """Test none for percents."""

        c = Color('color(--lch none 0 none)')
        self.assertEqual(c[1], 0)
        self.assertTrue(c.is_nan('l') and c.is_nan('h'))
        c = Color('hsl(30 none none)')
        self.assertEqual(c[0], 30)
        self.assertTrue(c.is_nan('s') and c.is_nan('l'))

    def test_normalize(self):
        """
        Test normalize.

        Should do nothing on a non-hue color.
        """

        self.assertColorEqual(Color('white').normalize(), Color('white'))
        self.assertColorNotEqual(Color('hsl(270 0% 50%)').normalize(), Color('hsl(270 0% 50%)'))

    def test_color_dict(self):
        """Color dictionaries."""

        c1 = Color('red')
        d = c1.to_dict()
        c2 = Color(d)

        self.assertEqual(c1, c2)

    def test_dict_input(self):
        """Test dictionary inputs."""

        c1 = Color('srgb', [1, 0, 1], 0.5)
        c2 = Color({'space': 'srgb', 'coords': [1, 0, 1], 'alpha': 0.5})
        c3 = Color({'space': 'srgb', 'coords': [1, 0, 1], 'alpha': 0.5})
        self.assertColorEqual(c1, c2)
        self.assertColorEqual(c1, c3)

    def test_white(self):
        """Test white."""

        self.assertEqual(Color('red').white(), [0.9504559270516716, 1, 1.0890577507598784])

    def test_less_input(self):
        """Test when not enough color channels are provided."""

        with self.assertRaises(ValueError):
            Color('color(srgb 1)')

    def test_non_ident(self):
        """Test color when its missing the identifier."""

        with self.assertRaises(ValueError):
            Color('color(1 1 1 / 0.5)')

    def test_missing_alpha(self):
        """Test missing alpha after slash."""

        with self.assertRaises(ValueError):
            Color('color(srgb 1 1 1 /)')

    def test_bad_color_channel_type(self):
        """Test bad color channel type."""

        with self.assertRaises(ValueError):
            Color('color(srgb 1deg 1 1)')

    def test_missing_end(self):
        """Test missing function end."""

        with self.assertRaises(ValueError):
            Color('color(srgb 1 1 1')

    def test_less_raw_input(self):
        """Test when not enough color channels are provided via raw input."""

        self.assertEqual(Color("srgb", [1]), Color("srgb", [1, NaN, NaN]))

    def test_too_many_input(self):
        """Test when too many color channels are provided."""

        with self.assertRaises(ValueError):
            Color("color(srgb 1 0 0 0 / 1)")

    def test_too_many_raw_input(self):
        """Test when too many color channels are provided via raw input."""

        with self.assertRaises(AttributeError):
            Color("srgb", [1, 0, 0, 0])

    def test_bad_input(self):
        """Test bad input."""

        with self.assertRaises(TypeError):
            Color(3)

    def test_bad_sytnax_input(self):
        """Test bad string syntax input."""

        with self.assertRaises(ValueError):
            Color("nope")

        with self.assertRaises(ValueError):
            Color("nope", [0, 0, 0])

    def test_bad_class(self):
        """Test bad class."""

        c = ColorAll('hunter-lab', [0, 0, 0])
        with self.assertRaises(ValueError):
            Color(c)

    def test_bad_data_input(self):
        """Test bad data input."""

        with self.assertRaises(ValueError):
            Color("srgb")

    def test_missing_values(self):
        """Test missing values."""

        with self.assertRaises(ValueError):
            Color('color(srgb)')

    def test_missing_inputs(self):
        """Test missing inputs."""

        coords = Color("srgb", [])[:-1]
        for c in coords:
            self.assertTrue(math.isnan(c))

    def test_new(self):
        """Test new."""

        c1 = Color('purple')
        c2 = c1.new('purple')
        self.assertEqual(c1, c2)

    def test_clone(self):
        """Test clone."""

        c1 = Color('purple')
        c2 = c1.clone()
        self.assertEqual(c1, c2)

    def test_update(self):
        """Test update."""

        c1 = Color('orange')
        c2 = Color('purple')
        self.assertNotEqual(c1, c2)
        c2.update(c1)
        self.assertEqual(c1, c2)

    def test_update_different_space(self):
        """Update from different space."""

        c1 = Color('orange')
        c2 = c1.convert('hsl')
        c3 = Color('purple').convert('hsl')
        self.assertEqual(c2, c3.update(c1))

    def test_update_self(self):
        """Update with self."""

        c1 = Color('orange')
        self.assertEqual(c1.update(c1), c1)

    def test_mutate(self):
        """Test mutate."""

        c1 = Color('orange').convert('lch')
        c2 = Color('orange')
        self.assertNotEqual(c1, c2)
        c2.mutate(c1)
        self.assertEqual(c1, c2)

    def test_convert(self):
        """Test convert."""

        c1 = Color('orange')
        c2 = c1.convert('hsl')
        self.assertColorEqual(c2, Color('hsl(39, 100%, 50%)'), precision=0)

    def test_convert_fit(self):
        """Test convert fit."""

        c1 = Color('color(srgb 2 -1 0)')
        self.assertFalse(c1.in_gamut())
        c2 = c1.convert("hsl", fit=True)
        self.assertTrue(c2.in_gamut())
        c3 = c1.convert('hsl').clone().fit()
        self.assertColorEqual(c2, c3)

    def test_convert_fit_clip(self):
        """Test convert fit."""

        c1 = Color('color(srgb 2 -1 0)')
        self.assertFalse(c1.in_gamut())
        c2 = c1.convert("hsl", fit="clip")
        self.assertTrue(c2.in_gamut())
        c3 = c1.convert('hsl').clone().fit(method="clip")
        self.assertColorEqual(c2, c3)
        c4 = c1.convert("hsl", fit=True)
        self.assertColorNotEqual(c2, c4)

    def test_property(self):
        """Test set."""

        c1 = Color('red')
        self.assertEqual(c1.get('green'), 0)
        c1.set('green', 0.5)
        self.assertEqual(c1.get('green'), 0.5)

    def test_get(self):
        """Test get."""

        c1 = Color('orange')
        self.assertEqual(c1.get("red"), 1.0)

    def test_get_no_nans(self):
        """Test get with no NaN."""

        self.assertEqual(Color('white').convert('lch').get('h', nans=False), 0.0)
        self.assertEqual(Color('white').convert('lch').get(['c', 'h'], nans=False), [0.0, 0.0])

    def test_space_get(self):
        """Test get with another space."""

        c1 = Color('orange')
        self.assertEqual(c1.get("hsl.lightness"), 0.5)

    def test_get_bad(self):
        """Test bad get."""

        c1 = Color('orange')

        with self.assertRaises(ValueError):
            c1.get("bad")

    def test_get_bad_chain(self):
        """Test bad get."""

        c1 = Color('orange')

        with self.assertRaises(ValueError):
            c1.get("hsl.hue.wrong")

    def test_set(self):
        """Test set."""

        c1 = Color('orange')
        c1.set("red", 0.5)
        self.assertEqual(c1.get("red"), 0.5)

    def test_set_undefined(self):
        """Test set."""

        def set_test(value):
            """Test that value is not NaN."""

            assert not math.isnan(value)
            return value

        c1 = Color('gray').convert('hsl')
        c1.set("hue", set_test, nans=False)

    def test_set_undefined_space(self):
        """Test set."""

        def set_test(value):
            """Test that value is not NaN."""

            assert not math.isnan(value)
            return value

        Color('gray').set('hsl.hue', set_test, nans=False)

    def test_multi_set_undefined(self):
        """Test set."""

        def set_test(value):
            """Test that value is not NaN."""

            assert not math.isnan(value)
            return value

        c1 = Color('gray').convert('hsl')
        c1.set({'lightness': 0.5, "hue": set_test}, nans=False)

    def test_multi_set_undefined_space(self):
        """Test set."""

        def set_test(value):
            """Test that value is not NaN."""

            assert not math.isnan(value)
            return value

        Color('gray').set({'hsl.saturation': 0.5, 'hsl.hue': set_test}, nans=False)

    def test_multi_set(self):
        """Test setting multiple channels via set."""

        color = Color('orange')
        color2 = color.clone()
        color2.convert('oklch', in_place=True)
        color2.set('hue', 270).set('lightness', lambda l: l - l * 0.25)
        color2.convert('srgb', in_place=True).set('alpha', 0.5)
        color.set(
            {
                'oklch.lightness': lambda l: l - l * 0.25,
                'alpha': 0.5,
                'oklch.hue': 270
            }
        )

        self.assertColorEqual(color, color2)

    def test_bad_multi_set_dict(self):
        """Test that a dictionary input with a value fails."""

        with self.assertRaises(ValueError):
            Color('red').set({'red': 0}, 0)

    def test_bad_multi_set_string(self):
        """Test that a string input with no value fails."""

        with self.assertRaises(ValueError):
            Color('red').set('red')

    def test_multi_get(self):
        """Test that we can get multiple values."""

        color = Color('orange')
        color.get(['oklch.lightness', 'alpha', 'oklch.hue'])
        oklch = color.convert('oklch')
        self.assertEqual(
            color.get(['oklch.lightness', 'alpha', 'oklch.hue']),
            [oklch['l'], color[-1], oklch['h']]
        )

    def test_space_set(self):
        """Test set in another space."""

        c1 = Color('orange')
        c1.set("hsl.hue", 270)
        self.assertEqual(c1.get("hsl.hue"), 270)

    def test_function_set(self):
        """Test set with a function."""

        c1 = Color('orange')
        c1.set("red", lambda x: x * 0.3)
        self.assertEqual(c1.get("red"), 0.3)

    def test_set_bad(self):
        """Test bad set."""

        c1 = Color('orange')

        with self.assertRaises(ValueError):
            c1.set("bad", 0.5)

    def test_set_bad_chain(self):
        """Test bad set."""

        c1 = Color('orange')

        with self.assertRaises(ValueError):
            c1.set("hsl.hue.wrong", 0.5)

    def test_set_bad_input(self):
        """Test bad set."""

        c1 = Color('orange')

        with self.assertRaises(ValueError):
            c1.set("red", "bad")

    def test_is_achromatic(self):
        """Test the default `is_achroamtic` (evaluation in XYZ-d65) when space reports it doesn't know."""

        from coloraide.spaces.srgb.css import sRGB

        class TempsRGB(sRGB):
            """Construct an sRGB class that defaults to fallback achromatic test."""

            def is_achromatic(self, coords):
                """
                If no undefined, report we don't know if the color is achromatic.

                This is the default and allows us to test the fallback achromatic test.
                """

                return None

        class TempColor(Color):
            """Temporary color class."""

        TempColor.register(TempsRGB(), overwrite=True)

        self.assertFalse(TempColor('red').is_achromatic())
        self.assertTrue(TempColor('gray').is_achromatic())

    def test_is_nan_false(self):
        """Test when `is_nan` is false."""

        self.assertFalse(Color('red').convert('hsl').is_nan('hue'))

    def test_is_nan_true(self):
        """Test when `is_nan` is true."""

        self.assertTrue(Color('white').convert('hsl').is_nan('hue'))

    def test_is_nan_false_different_space(self):
        """Test when `is_nan` is false."""

        self.assertFalse(Color('red').is_nan('hsl.hue'))

    def test_is_nan_true_different_space(self):
        """Test when `is_nan` is true."""

        self.assertTrue(Color('white').is_nan('hsl.hue'))

    def test_match(self):
        """Test match."""

        obj = Color.match('red')
        self.assertEqual(obj.color, Color('red'))
        self.assertEqual(obj.start, 0)
        self.assertEqual(obj.end, 3)

    def test_match_no_fullmatch(self):
        """Test match without full match."""

        obj = Color.match('red more text')
        self.assertEqual(obj.color, Color('red'))
        self.assertEqual(obj.start, 0)
        self.assertEqual(obj.end, 3)

    def test_match_fullmatch(self):
        """Test match with full match."""

        self.assertIsNone(Color.match('red more text', fullmatch=True))
        self.assertIsNotNone(Color.match('red', fullmatch=True))

    def test_match_offset(self):
        """Test match with offset."""

        obj = Color.match('yellow green #0000FF rgb(1, 0, 0)', start=7)
        self.assertEqual(obj.color, Color('green'))
        self.assertEqual(obj.start, 7)
        self.assertEqual(obj.end, 12)

        obj = Color.match('yellow green #0000FF rgb(1, 0, 0)', start=13)
        self.assertEqual(obj.color, Color('blue'))
        self.assertEqual(obj.start, 13)
        self.assertEqual(obj.end, 20)

        obj = Color.match('yellow green #0000FF rgb(255, 0, 0)', start=21)
        self.assertEqual(obj.color, Color('red'))
        self.assertEqual(obj.start, 21)
        self.assertEqual(obj.end, 35)

    def test_mask_in_place(self):
        """Test masking "in place"."""

        c1 = Color('white')
        c2 = c1.mask('red')
        self.assertNotEqual(c1, c2)
        self.assertIsNot(c1, c2)
        c3 = c1.mask('red', in_place=True)
        self.assertEqual(c1, c3)
        self.assertIs(c1, c3)

    def test_parse_float(self):
        """Test one that we generally handle floats with scientific notation."""

        self.assertColorEqual(Color("color(srgb 3.2e-2 0.1e+1 0.1e1 / 0.5)"), Color("color(srgb 0.032 1 1 / 0.5)"))
        self.assertColorEqual(Color("color(srgb +3.2e-2 +0.1e+1 +0.1e1 / 0.5)"), Color("color(srgb 0.032 1 1 / 0.5)"))

    def test_random_space(self):
        """Test that random colors are generated in the specified space."""

        c = Color.random('srgb')
        self.assertEqual(c.space(), 'srgb')

        c = Color.random('hsl')
        self.assertEqual(c.space(), 'hsl')

    def test_random_range(self):
        """Test that random colors are generated within the space's range."""

        for _ in range(10):
            for c in Color.random('srgb'):
                self.assertTrue(0 <= c <= 1)

    def test_random_limits(self):
        """Test random limits."""

        for _ in range(10):
            for i, c in enumerate(Color.random('srgb', limits=[None, (0, 0.5)])):
                if i == 1:
                    self.assertTrue(0 <= c <= 0.5)
                else:
                    self.assertTrue(0 <= c <= 1)



"""
Rec 2100 HLG color class.

https://www.itu.int/dms_pubrec/itu-r/rec/bt/R-REC-BT.2100-2-201807-I!!PDF-E.pdf

https://www.itu.int/dms_pub/itu-r/opb/rep/R-REP-BT.2390-10-2021-PDF-E.pdf (page 25)

> Report ITU-R BT.2408 indicates that, for HLG HDR, diffuse white should be set at a signal level of
> 75%. This can be configured by making the output from an 18% grey card correspond to a signal
> level of 38%, rather than the 42.5% stated above.

https://lists.w3.org/Archives/Public/public-colorweb/2021Sep/0008.html

Suggests the scale of 0.26496256042100724 to satisfy the above requirement.

"""
from __future__ import annotations
from ..cat import WHITES
from .srgb import sRGB
from .. import algebra as alg
from ..types import Vector
import math

# Sets 18% grey card to ~38% (37.7...) in order to set diffused white to 75%.
SCALE = 0.26496256042100724


class Environment:
    """Class to calculate and contain any required environmental data."""

    def __init__(
        self,
        lw: float,
        lb: float,
        scale: float
    ):
        """Initialize environmental data."""

        self.a = 0.17883277
        self.b = 0.28466892  # `1 - 4 * a`
        self.c = 0.55991073  # `0.5 - a * math.log(4 * a)`
        self.beta = hlg_black_level_lift(lw, lb)
        self.scale = scale
        self.inv_scale = 1 / scale


def hlg_gamma(lw: float = 1000.0) -> float:
    """
    Return the reference HLG system gamma.

    Ranges should be `lw >= 1000 cd / m^2`.
    """

    return 1.2 + 0.42 * math.log(lw / 1000.0)


def hlg_black_level_lift(lw: float = 0.0, lb: float = 1000.0) -> float:
    """
    Return beta (the black level lift) using the nominal peak level luminance and display luminance for black.

    Ranges should be `lw >= 1000 cd / m^2` and `lb <= 0.005 cd / m^2`.
    """

    return math.sqrt(3 * (lb / lw) ** (1 / hlg_gamma(lw)))


def hlg_oetf(values: Vector, env: Environment) -> Vector:
    """HLG OETF."""

    adjusted = []  # type: Vector
    for e in values:
        e = alg.nth_root(3 * e, 2) if e <= 1 / 12 else env.a * math.log(12 * e - env.b) + env.c
        adjusted.append((e - env.beta) / (1 - env.beta))
    return adjusted


def hlg_eotf(values: Vector, env: Environment) -> Vector:
    """HLG EOTF."""

    adjusted = []  # type: Vector
    for e in values:
        e = (1 - env.beta) * e + env.beta
        adjusted.append((e ** 2) / 3 if e <= 0.5 else (math.exp((e - env.c) / env.a) + env.b) / 12)
    return adjusted


class Rec2100HLG(sRGB):
    """Rec. 2100 HLG class."""

    BASE = "rec2020-linear"
    NAME = "rec2100-hlg"
    SERIALIZE = ('--rec2100-hlg',)
    WHITE = WHITES['2deg']['D65']
    DYNAMIC_RANGE = 'hdr'
    ENV = Environment(1000, 0, SCALE)

    def to_base(self, coords: Vector) -> Vector:
        """To base from Rec 2100 HLG."""

        return [c * self.ENV.inv_scale for c in hlg_eotf(coords, self.ENV)]

    def from_base(self, coords: Vector) -> Vector:
        """From base to Rec. 2100 HLG."""

        return hlg_oetf([c * self.ENV.scale for c in coords], self.ENV)



"""HyAB distance."""
from __future__ import annotations
from ..distance import DeltaE
import math
from ..spaces import Labish
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class DEHyAB(DeltaE):
    """Delta E HyAB class."""

    NAME = "hyab"

    def __init__(self, space: str = "lab-d65") -> None:
        """Initialize."""

        self.space = space

    def distance(self, color: Color, sample: Color, space: str | None = None, **kwargs: Any) -> float:
        """
        HyAB distance for Lab-ish spaces.

        http://markfairchild.org/PDFs/PAP40.pdf.
        """

        if space is None:
            space = self.space

        color = color.convert(space)
        sample = sample.convert(space)

        if not isinstance(color._space, Labish):
            raise ValueError("The space '{}' is not a 'lab-ish' color space and cannot use HyAB".format(space))

        names = color._space.names()
        l1, a1, b1 = color.get(names, nans=False)
        l2, a2, b2 = sample.get(names, nans=False)

        return abs(l1 - l2) + math.sqrt((a1 - a2) ** 2 + (b1 - b2) ** 2)



"""
Interpolation methods.

Originally, the base code for `interpolate`, `mix` and `steps` was ported from the
https://colorjs.io project. Since that time, there has been significant modifications
that add additional features etc. The base logic though is attributed to the original
authors.

In general, the logic mimics in many ways the `color-mix` function as outlined in the Level 5
color draft (Oct 2020), but the initial approach was modeled directly off of the work done in
color.js.
---
Original Authors: Lea Verou, Chris Lilley
License: MIT (As noted in https://github.com/LeaVerou/color.js/blob/master/package.json)
"""
from __future__ import annotations
import math
import functools
from abc import ABCMeta, abstractmethod
from .. import util
from .. import algebra as alg
from .. spaces import HSVish, HSLish, Cylindrical, RGBish, LChish, Labish
from ..types import Vector, ColorInput, Plugin
from typing import Callable, Sequence, Mapping, Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color

__all__ = ('stop', 'hint', 'get_interpolator')


class stop:
    """Color stop."""

    __slots__ = ('color', 'stop')

    def __init__(self, color: ColorInput, value: float) -> None:
        """Color stops."""

        self.color = color
        self.stop = value


def midpoint(t: float, h: float = 0.5) -> float:
    """Midpoint easing function."""

    return 0.0 if h <= 0 or h >= 1 else alg.npow(t, math.log(0.5) / math.log(h))


def hint(mid: float) -> Callable[..., float]:
    """A generate a midpoint easing function."""

    return functools.partial(midpoint, h=mid)


def normalize_domain(d: list[float]) -> list[float]:
    """Normalize domain between 0 and 1."""

    total = d[-1] - d[0]
    regions = len(d) - 1
    values = [0.0]
    for index in range(regions):
        a, b = d[index:index + 2]
        l = b - a
        values.append(values[-1] + (l / total if total else 0))
    return values


class Interpolator(metaclass=ABCMeta):
    """Interpolator."""

    def __init__(
        self,
        coordinates: list[Vector],
        channel_names: Sequence[str],
        create: type[Color],
        easings: list[Callable[..., float] | None],
        stops: dict[int, float],
        space: str,
        out_space: str,
        progress: Mapping[str, Callable[..., float]] | Callable[..., float] | None,
        premultiplied: bool,
        extrapolate: bool = False,
        domain: Sequence[float] | None = None,
        padding: float | tuple[float, float] | None = None,
        **kwargs: Any
    ):
        """Initialize."""

        self.start = stops[0]
        self.end = stops[len(stops) - 1]
        self.stops = stops
        self.easings = easings
        self.coordinates = coordinates
        self.length = len(self.coordinates)
        self.channel_names = channel_names
        self.create = create
        self.progress = progress
        self.space = space
        self._out_space = out_space
        self.extrapolate = extrapolate
        self.current_easing = None  # type: Mapping[str, Callable[..., float]] | Callable[..., float] | None
        cs = self.create.CS_MAP[space]
        if isinstance(cs, Cylindrical):
            self.hue_index = cs.hue_index()
        else:
            self.hue_index = -1
        self.premultiplied = premultiplied

        # Calculate padded start and end
        self._padding = None  # type: tuple[float, float] | None
        if padding is not None:
            self.padding(padding)

        # Set the domain
        self._domain = []  # type: list[float]
        if domain is not None:
            self.domain(domain)

        self.setup()

    def discretize(
        self,
        steps: int = 2,
        max_steps: int = 1000,
        max_delta_e: float = 0,
        delta_e: str | None = None
    ) -> None:
        """Make the interpolation a discretized interpolation."""

        # Get the discrete steps for the new discrete interpolation
        colors = self.steps(steps, max_steps, max_delta_e, delta_e)

        # Calculate new coordinate list and discrete stops
        total = len(colors)
        coords = []
        stops = {}
        count = 0
        for r in range(1, total):
            pre = r - 1
            nxt = r
            step1 = colors[pre][:]
            step2 = colors[nxt][:]
            stp = r / total
            stops[count] = stp
            stops[count + 1] = stp
            coords.extend([step1, step2])
            count += 2

        # Update colors and stops
        self.coordinates = coords
        self.length = len(self.coordinates)
        self.stops = stops
        self.start = self.stops[0]
        self.end = self.stops[len(self.stops) - 1]

        # Reset features that were used to generate the discrete steps
        self.easings = [None] * (self.length - 1)
        self.progress = None
        self.current_easing = None
        self._padding = None
        self._domain = []

        self.setup()

    def out_space(self, space: str) -> None:
        """Set output space."""

        if space not in self.create.CS_MAP:
            raise ValueError("'{}' is not a valid color space".format(space))
        self._out_space = space

    def padding(self, padding: float | Sequence[float]) -> None:
        """Add/adjust padding."""

        # Make sure it is a sequence
        padding = [padding] if not isinstance(padding, Sequence) else list(padding)

        # If it is empty
        if not padding:
            self._padding = None
            return

        l = len(padding)

        # Too many values
        if l > 2:
            raise ValueError("Padding must be either a single numerical value or a sequence of 2 values")

        # Apply padding to both
        if l == 1:
            padding.append(padding[0])

        # No padding is required
        if padding[0] == padding[1] == 0.0:
            self._padding = None

        # Calculate padded start and end
        else:
            self._padding = (0.0 + padding[0], 1.0 - padding[1])

    def domain(self, domain: Sequence[float]) -> None:
        """Set the domain."""

        # Ensure domain ascends.
        # If we have a domain of length 1, we will duplicate it.
        d = []  # type: list[float]
        if domain:
            length = len(domain)

            # Ensure values are not descending
            d.append(domain[0])
            for index in range(length - 1):
                b = domain[index + 1]
                d.append(d[-1] if b <= d[-1] else b)

            # We need at least two values, so duplicate the first.
            if len(d) == 1:
                d.append(d[0])
            domain = d

        self._domain = d

    @abstractmethod
    def setup(self) -> None:
        """Setup."""

    @abstractmethod
    def interpolate(
        self,
        point: float,
        index: int,
    ) -> Vector:
        """Interpolate."""

    def steps(
        self,
        steps: int = 2,
        max_steps: int = 1000,
        max_delta_e: float = 0,
        delta_e: str | None = None
    ) -> list[Color]:
        """Steps."""

        actual_steps = steps

        # Allocate at least two steps if we are doing a maximum delta E,
        if max_delta_e != 0 and actual_steps < 2:
            actual_steps = 2

        # Make sure we don't start out allocating too many colors
        if max_steps is not None:
            actual_steps = min(actual_steps, max_steps)

        ret = []  # type: list[tuple[float, Color]]
        if actual_steps == 1:
            ret = [(0.5, self(0.5))]
        elif actual_steps > 1:
            step = 1 / (actual_steps - 1)
            for i in range(actual_steps):
                p = i * step
                ret.append((p, self(p)))

        # Iterate over all the stops inserting stops in between all colors
        # if we have any two colors with a max delta greater than what was requested.
        # We inject between every stop to ensure the midpoint does not shift.
        if max_delta_e > 0:
            # Initial check to see if we need to insert more stops
            m_delta = 0.0
            for i in range(1, len(ret)):
                m_delta = max(
                    m_delta,
                    ret[i - 1][1].delta_e(
                        ret[i][1],
                        method=delta_e
                    )
                )

            # If we currently have delta over our limit inject more stops.
            # If inserting between every color would push us over the max_steps, halt.
            total = len(ret)
            while m_delta > max_delta_e and (total * 2 - 1 <= max_steps):
                # Inject stops while measuring again to see if it was sufficient
                m_delta = 0.0
                i = 1
                index = 1
                while index < total:
                    prev = ret[index - 1]
                    cur = ret[index]
                    p = (cur[0] + prev[0]) / 2
                    color = self(p)
                    m_delta = max(
                        m_delta,
                        color.delta_e(prev[1], method=delta_e),
                        color.delta_e(cur[1], method=delta_e)
                    )
                    ret.insert(index, (p, color))
                    total += 1
                    index += 2

        return [i[1] for i in ret]

    def premultiply(self, coords: Vector, alpha: float | None = None) -> None:

        if alpha is not None:
            coords[-1] = alpha
        else:
            alpha = coords[-1]

        if alg.is_nan(alpha) or alpha == 1.0:
            return

        for i, value in enumerate(coords[:-1]):

            # Wrap the angle
            if i == self.hue_index:
                continue

            coords[i] = value * alpha

    def postdivide(self, coords: Vector) -> None:
        """Undo premultiplication of semi-transparent colors."""

        alpha = coords[-1]

        if alg.is_nan(alpha) or alpha in (0.0, 1.0):
            return

        for i, value in enumerate(coords[:-1]):

            # Wrap the angle
            if i == self.hue_index:
                continue

            coords[i] = value / alpha

    def begin(self, point: float, first: float, last: float, index: int) -> Color:
        """
        Begin interpolation.

        - Ensure point is relative to the stops.
        - Get the appropriate easing function.
        - Call interpolation.
        - Return a color
        """

        # Adjust stop to be relative to the given stops
        r = last - first
        if point < first:
            adjusted_time = point - first if self.extrapolate else 0
        elif point > last:
            adjusted_time = 1 + point - last if self.extrapolate else 1
        else:
            adjusted_time = (point - first) / r if r else 1

        # Do we have an easing function between these stops?
        self.current_easing = self.easings[index - 1]
        if self.current_easing is None:
            self.current_easing = self.progress

        # Interpolate color and return it
        coords = self.interpolate(adjusted_time, index)
        if self.premultiplied:
            self.postdivide(coords)

        # Create the color and ensure it is in the correct color space.
        color = self.create(self.space, coords[:-1], coords[-1])
        return color.convert(self._out_space, in_place=True)

    def ease(self, t: float, channel_index: int) -> float:
        """Provide a progression time and channel index."""

        progress = None
        if self.current_easing is not None:
            # Do we have an easing function, or mapping with a channel easing function?
            name = self.channel_names[channel_index]
            if isinstance(self.current_easing, Mapping):
                progress = self.current_easing.get(name)
                if progress is None:
                    progress = self.current_easing.get('all')
            else:
                progress = self.current_easing

        return progress(t) if progress is not None else t

    def scale(self, point: float) -> float:
        """
        Scale a point from a custom domain into a domain of 0 to 1.

        This allows a user to have a custom domain, but for us to adapt back to 0 and 1
        so that our logic can remain consistent.
        """

        if point < self._domain[0]:
            point = (point - self._domain[0]) / (self._domain[-1] - self._domain[0]) if self.extrapolate else 0.0
        elif point > self._domain[-1]:
            point = 1.0 + (point - self._domain[-1]) / (self._domain[-1] - self._domain[0]) if self.extrapolate else 1.0
        else:
            regions = len(self._domain) - 1
            size = (1 / regions)
            index = 0
            adjusted = 0.0
            for index in range(regions):
                a, b = self._domain[index:index + 2]
                if point >= a and point <= b:
                    l = b - a
                    adjusted = ((point - a) / l) if l else 0.0
                    break

            point = size * index + (adjusted * size)
        return point

    def __call__(self, point: float) -> Color:
        """Find which leg of the interpolation the request is between."""

        if self._domain:
            point = self.scale(point)

        if self._padding:
            slope = (self._padding[1] - self._padding[0])
            point = self._padding[0] + slope * point
            if not self.extrapolate:
                point = min(max(point, self._padding[0]), self._padding[1])

        # See if point extends past either the first or last stop
        if point < self.start:
            first, last = self.start, self.stops[1]
            return self.begin(point, first, last, 1)
        elif point > self.end:
            first, last = self.stops[self.length - 2], self.end
            return self.begin(point, first, last, self.length - 1)
        else:
            # Iterate stops to find where our point falls between
            first = self.start
            for i in range(1, self.length):
                last = self.stops[i]
                if point <= last:
                    return self.begin(point, first, last, i)
                first = last

        # We shouldn't ever hit this, but provided for typing.
        # If we do hit this, it would be a bug.
        raise RuntimeError('Iterpolation could not be found for {}'.format(point))  # pragma: no cover


class Interpolate(Plugin, metaclass=ABCMeta):
    """Interpolation plugin."""

    NAME = ""

    @abstractmethod
    def interpolator(
        self,
        coordinates: list[Vector],
        channel_names: Sequence[str],
        create: type[Color],
        easings: list[Callable[..., float] | None],
        stops: dict[int, float],
        space: str,
        out_space: str,
        progress: Mapping[str, Callable[..., float]] | Callable[..., float] | None,
        premultiplied: bool,
        extrapolate: bool = False,
        domain: list[float] | None = None,
        padding: float | tuple[float, float] | None = None,
        **kwargs: Any
    ) -> Interpolator:
        """Get the interpolator object."""


def calc_stops(stops: dict[int, float], count: int) -> dict[int, float]:
    """Calculate stops."""

    # Ensure the first stop is set to zero if not explicitly set
    if 0 not in stops or stops[0] is None:
        stops[0] = 0

    last = stops[0] * 100
    highest = last
    empty = None
    final = {}

    # Build up normalized stops
    for i in range(count):
        value = stops.get(i)
        if value is not None:
            value *= 100

        # Found an empty hole, track the start
        if value is None and empty is None:
            empty = i - 1
            continue
        elif value is None:
            continue

        # We can't have a stop decrease in progression
        if value < last:
            value = last

        # Track the largest explicit value set
        if value > highest:
            highest = value

        # Fill in hole if one exists.
        # Holes will be evenly space between the
        # current and last stop.
        if empty is not None:
            r = i - empty
            increment = (value - last) / r
            for j in range(empty + 1, i):
                last += increment
                final[j] = last / 100
            empty = None

        # Set the stop and track it as the last
        last = value
        final[i] = last / 100

    # If there is a hole at the end, fill in the hole,
    # equally spacing the stops from the last to 100%.
    # If the last is greater than 100%, then all will
    # be equal to the last.
    if empty is not None:
        r = (count - 1) - empty
        if highest > 100:
            increment = 0
        else:
            increment = (100 - last) / r
        for j in range(empty + 1, count):
            last += increment
            final[j] = last / 100

    return final


def process_mapping(
    progress: Mapping[str, Callable[..., float]] | Callable[..., float] | None,
    aliases: Mapping[str, str]
) -> Mapping[str, Callable[..., float]] | Callable[..., float] | None:
    """Process a mapping, such that it is not using aliases."""

    if not isinstance(progress, Mapping):
        return progress
    return {aliases.get(k, k): v for k, v in progress.items()}


def adjust_shorter(h1: float, h2: float, offset: float) -> tuple[float, float]:
    """Adjust the given hues."""

    d = h2 - h1
    if d > 180:
        h2 -= 360.0
        offset -= 360.0
    elif d < -180:
        h2 += 360
        offset += 360.0
    return h2, offset


def adjust_longer(h1: float, h2: float, offset: float) -> tuple[float, float]:
    """Adjust the given hues."""

    d = h2 - h1
    if 0 < d < 180:
        h2 -= 360.0
        offset -= 360.0
    elif -180 < d <= 0:
        h2 += 360
        offset += 360.0
    return h2, offset


def adjust_increase(h1: float, h2: float, offset: float) -> tuple[float, float]:
    """Adjust the given hues."""

    if h2 < h1:
        h2 += 360.0
        offset += 360.0
    return h2, offset


def adjust_decrease(h1: float, h2: float, offset: float) -> tuple[float, float]:
    """Adjust the given hues."""

    if h2 > h1:
        h2 -= 360.0
        offset -= 360.0
    return h2, offset


def normalize_hue(
    color1: Vector,
    color2: Vector | None,
    index: int,
    offset: float,
    hue: str,
    fallback: float | None
) -> tuple[Vector, float]:
    """Normalize hues according the hue specifier."""

    if hue == 'specified':
        return (color2 or color1), offset

    # Probably the first hue
    if color2 is None:
        color1[index] = util.constrain_hue(color1[index])
        return color1, offset

    if hue == 'shorter':
        adjuster = adjust_shorter
    elif hue == 'longer':
        adjuster = adjust_longer
    elif hue == 'increasing':
        adjuster = adjust_increase
    elif hue == 'decreasing':
        adjuster = adjust_decrease
    else:
        raise ValueError("Unknown hue adjuster '{}'".format(hue))

    c1 = color1[index] + offset
    c2 = util.constrain_hue(color2[index]) + offset

    # Adjust hue, handle gaps across `NaN`s
    if not alg.is_nan(c2):
        if not alg.is_nan(c1):
            c2, offset = adjuster(c1, c2, offset)
        elif fallback is not None:
            c2, offset = adjuster(fallback, c2, offset)

    color2[index] = c2
    return color2, offset


def carryforward_convert(color: Color, space: str) -> None:  # pragma: no cover
    """Carry forward undefined values during conversion."""

    cs1 = color._space
    cs2 = color.CS_MAP[space]
    channels = {'r': False, 'g': False, 'b': False, 'h': False, 'c': False, 'l': False, 'v': False}
    carry = []

    # Gather undefined channels
    if isinstance(cs1, RGBish):
        for i, name in zip(cs1.indexes(), ('r', 'g', 'b')):
            if alg.is_nan(color[i]):
                channels[name] = True
    elif isinstance(cs1, LChish):
        for i, name in zip(cs1.indexes(), ('l', 'c', 'h')):
            if alg.is_nan(color[i]):
                channels[name] = True
    elif isinstance(cs1, Labish):
        if alg.is_nan(color[cs1.indexes()[0]]):
            channels['l'] = True
    elif isinstance(cs1, HSLish):
        for i, name in zip(cs1.indexes(), ('h', 'c', 'l')):
            if alg.is_nan(color[i]):
                channels[name] = True
    elif isinstance(cs1, HSVish):
        for i, name in zip(cs1.indexes(), ('h', 'c', 'v')):
            if alg.is_nan(color[i]):
                channels[name] = True
    elif isinstance(cs1, Cylindrical):
        if alg.is_nan(color[cs1.hue_index()]):
            channels['h'] = True

    # Carry alpha forward if undefined
    if alg.is_nan(color[-1]):
        carry.append(-1)

    # Channels that need to be carried forward
    if isinstance(cs2, RGBish):
        indexes = cs2.indexes()
        for e, name in enumerate(('r', 'g', 'b')):
            if channels[name]:
                carry.append(indexes[e])
    elif isinstance(cs2, Labish):
        indexes = cs2.indexes()
        if channels['l']:
            carry.append(indexes[0])
    elif isinstance(cs2, LChish):
        indexes = cs2.indexes()
        for e, name in enumerate(('l', 'c', 'h')):
            if channels[name]:
                carry.append(indexes[e])
    elif isinstance(cs2, HSLish):
        indexes = cs2.indexes()
        for e, name in enumerate(('h', 'c', 'l')):
            if channels[name]:
                carry.append(indexes[e])
    elif isinstance(cs2, HSVish):
        indexes = cs2.indexes()
        for e, name in enumerate(('h', 'c', 'v')):
            if channels[name]:
                carry.append(indexes[e])
    elif isinstance(cs2, Cylindrical):
        if channels['h']:
            carry.append(cs2.hue_index())

    # Convert the color space
    color.convert(space, in_place=True)
    for i in carry:
        color[i] = alg.NaN


def interpolator(
    interpolator: str,
    create: type[Color],
    colors: Sequence[ColorInput | stop | Callable[..., float]],
    space: str | None,
    out_space: str | None,
    progress: Mapping[str, Callable[..., float]] | Callable[..., float] | None,
    hue: str,
    premultiplied: bool,
    extrapolate: bool,
    domain: list[float] | None = None,
    padding: float | tuple[float, float] | None = None,
    **kwargs: Any
) -> Interpolator:
    """Get desired blend mode."""

    plugin = create.INTERPOLATE_MAP.get(interpolator)
    if not plugin:
        raise ValueError("'{}' is not a recognized interpolator".format(interpolator))

    # Construct piecewise interpolation object
    stops = {}  # type: Any

    if space is None:
        space = create.INTERPOLATE

    if isinstance(colors[0], stop):
        current = create(colors[0].color)
        stops[0] = colors[0].stop
    elif not callable(colors[0]):
        current = create(colors[0])
        stops[0] = None
    else:
        raise ValueError('Cannot have an easing function as the first item in an interpolation list')

    if out_space is None:
        out_space = space

    # Adjust to space
    if space != current.space():
        if kwargs.get('_carryforward', False):  # pragma: no cover
            carryforward_convert(current, space)
        else:
            current.convert(space, in_place=True)

    offset = 0.0
    hue_index = current._space.hue_index() if isinstance(current._space, Cylindrical) else -1

    # Normalize hue
    norm_coords = current[:]
    fallback = None
    if hue_index >= 0:
        h = norm_coords[hue_index]
        norm_coords, offset = normalize_hue(norm_coords, None, hue_index, offset, hue, fallback)
        if not alg.is_nan(h):
            fallback = h

    easing = None  # type: Any
    easings = []  # type: Any
    coords = [norm_coords]

    i = 0
    for x in colors[1:]:

        # Normalize all colors as Piecewise objects
        if isinstance(x, stop):
            i += 1
            stops[i] = x.stop
            color = current.new(x.color)
        elif callable(x):
            easing = x
            continue
        else:
            i += 1
            color = current.new(x)
            stops[i] = None

        # Adjust color to space
        if space != color.space():
            if kwargs.get('_carryforward', False):  # pragma: no cover
                carryforward_convert(color, space)
            else:
                color.convert(space, in_place=True)

        # Normalize the hue
        norm_coords = color[:]
        if hue_index >= 0:
            h = norm_coords[hue_index]
            norm_coords, offset = normalize_hue(current[:], norm_coords, hue_index, offset, hue, fallback)
            if not alg.is_nan(h):
                fallback = h

        # Create an entry interpolating the current color and the next color
        coords.append(norm_coords)
        easings.append(easing if easing is not None else progress)

        # The "next" color is now the "current" color
        easing = None
        current = color

    i += 1
    if i < 2:
        raise ValueError('Need at least two colors to interpolate')

    # Calculate stops
    stops = calc_stops(stops, i)

    # Send the interpolation list along with the stop map to the Piecewise interpolator
    return plugin.interpolator(
        coords,
        current._space.channels,
        create,
        easings,
        stops,
        space,
        out_space,
        process_mapping(progress, current._space.CHANNEL_ALIASES),
        premultiplied,
        extrapolate,
        domain,
        padding,
        **kwargs
    )
